# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains helper function to extract code from the .md files."""
import os
import re
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Callable, Dict, List, Optional

import click
import mistune  # type: ignore

from autonomy.cli.core import autonomy_cli

from tests.conftest import ROOT_DIR


MISTUNE_BLOCK_CODE_ID = "block_code"
IPFS_HASH_REGEX = r"Qm[A-Za-z0-9]{44}"
PYTHON_LINE_COMMENT_REGEX = r"^#.*\n"
DOC_ELLIPSIS_REGEX = r"\s*#\s...\n"
PYTHON_COMMAND = r"^pyt(hon|est) (?P<file_name>.*\.py).*$"
MAKE_COMMAND = r"^make (?P<cmd_name>.*)$"
AUTONOMY_COMMAND = r"^(?P<cmd_name>autonomy .*)$"
MAKEFILE_COMMAND = r"^(?P<command>.*):$"


class CodeType(Enum):
    """Types of code blocks in the docs"""

    PYTHON = "python"
    YAML = "yaml"
    BASH = "bash"
    NOCODE = "nocode"


def block_code_filter(b: Dict) -> bool:
    """Check Mistune block is a code block."""
    return b["type"] == MISTUNE_BLOCK_CODE_ID


def type_filter(type_: Optional[str], b: Dict) -> bool:
    """
    Check Mistune code block is of a certain type.

    If the field "info" is None, return False.
    If type_ is None, this function always return true.

    :param type_: the expected type of block (optional)
    :param b: the block dicionary.
    :return: True if the block should be accepted, false otherwise.
    """
    if type_ is None:
        return True
    return b["info"].strip() == type_ if b["info"] is not None else False


def extract_code_blocks(filepath: str, filter_: Optional[str] = None) -> list:
    """Extract code blocks from .md files."""
    content = Path(filepath).read_text(encoding="utf-8")
    markdown_parser = mistune.create_markdown(renderer=mistune.AstRenderer())
    blocks = markdown_parser(content)
    actual_type_filter = partial(type_filter, filter_)
    code_blocks = list(filter(block_code_filter, blocks))
    bash_code_blocks = filter(actual_type_filter, code_blocks)
    return list(b["text"] for b in bash_code_blocks)


def extract_python_code(filepath: str) -> str:
    """Removes the license part from the scripts"""
    python_str = ""
    with open(filepath, "r") as python_file:
        read_python_file = python_file.readlines()
    for i in range(21, len(read_python_file)):
        python_str += read_python_file[i]

    return python_str


def read_file(filepath: str) -> str:
    """Loads a file into a string"""
    with open(filepath, "r") as file_:
        file_str = file_.read()
    return file_str


def remove_line_comments(string: str) -> str:
    """Removes tokens from a python string"""
    return re.sub(PYTHON_LINE_COMMENT_REGEX, "", string)


def remove_doc_ellipsis(string: str) -> str:
    """Removes # ... from a python string"""
    return re.sub(DOC_ELLIPSIS_REGEX, "", string)


def remove_ips_hashes(string: str) -> str:
    """Replaces IPFS hashes with a placeholder"""
    return re.sub(IPFS_HASH_REGEX, "<ipfs_hash>", string)


def contains_code_blocks(file_path: str, block_type: str) -> bool:
    """Check if a doc file contains code blocks"""
    doc_path = os.path.join(ROOT_DIR, file_path)
    code_blocks = extract_code_blocks(filepath=doc_path, filter_=block_type)
    return len(code_blocks) > 0


def check_code_blocks_exist(
    md_file: str,
    code_info: Dict,
    code_type: CodeType,
    doc_process_fn: Optional[Callable] = None,
    code_process_fn: Optional[Callable] = None,
) -> None:
    """Check code blocks from the documentation"""
    code_files: List = code_info.get("code_files", None)
    skip_blocks: List = code_info.get("skip_blocks", None)

    # Load the code blocks from the doc file
    doc_path = os.path.join(ROOT_DIR, md_file)
    code_blocks = extract_code_blocks(filepath=doc_path, filter_=code_type.value)

    if skip_blocks:
        code_blocks = [
            code_blocks[i] for i in range(len(code_blocks)) if i not in skip_blocks
        ]

    if not code_blocks:
        return

    # Process the code blocks
    code_blocks = (
        list(map(doc_process_fn, code_blocks)) if doc_process_fn else code_blocks
    )

    # Ensure the code block mapping is correct. We ned a code file for each code block in the doc file
    assert len(code_blocks) == len(
        code_files
    ), f"Doc checker found {len(code_blocks)} non-skipped code blocks in {md_file} but only {len(code_files)} are being checked"

    for i, code_file in enumerate(code_files):
        # Check if the match must be performed line by line
        line_by_line = code_file.startswith("by_line::")
        code_file = code_file.replace("by_line::", "")

        # Load the code file and process it
        code_path = os.path.join(ROOT_DIR, code_file)
        code = read_file(code_path)
        code = code_process_fn(code) if code_process_fn else code

        # Perform the check
        if line_by_line:
            for line in code_blocks[i].split("\n"):
                assert (
                    line in code
                ), f"This line in {md_file} doesn't exist in the code file {code_file}:\n\n'{line}'"
            continue
        assert (
            code_blocks[i] in code
        ), f"This code-block in {md_file} doesn't exist in the code file {code_file}:\n\n{code_blocks[i]}"


def extract_make_commands(makefile_paths: List[str]) -> List[str]:
    """Extract make commands from a file"""
    commands = ["clean"]
    for makefile_path in makefile_paths:
        with open(makefile_path, "r", encoding="utf-8") as makefile:
            for line in makefile.readlines():
                match = re.match(MAKEFILE_COMMAND, line)
                if match:
                    commands += match.groupdict()["command"].split(" ")
    return commands


def extract_autonomy_commands() -> List[str]:
    """Extract autonomy commands from the autonomy cli"""
    cmd_list = []
    for cmd_name in autonomy_cli.list_commands(click.Context):
        cmd = autonomy_cli.get_command(click.Context, cmd_name)
        cmd_list += [
            f"autonomy {cmd_name} {sub_cmd_name}"
            for sub_cmd_name in cmd.list_commands(click.Context)
        ]
    return cmd_list


def check_bash_commands_exist(
    md_file: str, make_commands: List[str], autonomy_commands: List[str]
) -> None:
    """Check whether a bash code block exists in the codebase"""
    # Load the code file and process it
    doc_path = os.path.join(ROOT_DIR, md_file)
    code_blocks = extract_code_blocks(filepath=doc_path, filter_=CodeType.BASH.value)

    for code_block in code_blocks:
        for line in code_block.split("\n"):

            # Python/pytest commands
            match = re.match(PYTHON_COMMAND, line)
            if match:
                file_name = os.path.join(
                    ROOT_DIR, ROOT_DIR, match.groupdict()["file_name"]
                )
                assert os.path.isfile(
                    file_name
                ), f"File {file_name} referenced in {md_file} does not exist"
                continue

            # Make commands
            match = re.match(MAKE_COMMAND, line)
            if match:
                mk_cmds = match.groupdict()["cmd_name"]
                for mk_cmd in mk_cmds.split(" "):
                    assert (
                        mk_cmd in make_commands
                    ), f"Make command '{mk_cmd}' referenced in {md_file} is not present in the Makefile"
                continue

            # autonomy commands
            match = re.match(AUTONOMY_COMMAND, line)
            if match:
                autonomy_cmd = match.groupdict()["cmd_name"]
                assert any(
                    autonomy_cmd.startswith(cmd) for cmd in autonomy_commands
                ), f"autonomy command '{autonomy_cmd}' referenced in {md_file} is not present in the autonomy cli"
                continue



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Test to ensure that tendermint classes used for deployment stay in sync"""

import difflib
import inspect
import logging
from typing import List, Type

import pytest

from deployments.Dockerfiles.localnode import tendermint  # type: ignore

from packages.valory.connections.abci import connection


@pytest.mark.parametrize(
    "deployment_cls, package_cls",
    [
        (connection.StoppableThread, tendermint.StoppableThread),
        (connection.TendermintParams, tendermint.TendermintParams),
        (connection.TendermintNode, tendermint.TendermintNode),
    ],
)
def test_deployment_class_identical(deployment_cls: Type, package_cls: Type) -> None:
    """Assert Tendermint deployment and package code is identical"""

    def get_lines(cls: Type) -> List[str]:
        return inspect.getsource(cls).splitlines(keepends=False)

    p_code, d_code = map(get_lines, (deployment_cls, package_cls))
    differences = "\n".join(difflib.unified_diff(p_code, d_code, lineterm=""))
    if differences:
        logging.error(differences)
    assert not differences



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test optimization operations."""
from typing import Any, Callable
from unittest.mock import MagicMock

import numpy as np
import optuna
import pandas as pd
import pmdarima.model_selection
import pytest
from _pytest.monkeypatch import MonkeyPatch
from sklearn.metrics import mean_pinball_loss

from packages.valory.skills.apy_estimation_abci.ml.optimization import (
    Objective,
    get_best_params,
    optimize,
    optimize_single_pool,
    pinball_loss_scorer,
)


class TestOptimization:
    """Tests for the optimization operations."""

    @staticmethod
    @pytest.mark.parametrize(
        "y_true,y_pred,alpha",
        (
            (
                np.asarray([4.45, 4.434, 8.45, 0.342, 7.34, 6346.3475, 3.0]),
                np.asarray([4.45, 5.434, 34.23, 0.345634, 2.34, 633456.3475, 3.0]),
                0.1,
            ),
            (
                np.asarray([6.45, 3.434, 7.45, 0.45342, 37.34, 0.3475, 36]),
                np.asarray([445, 5434, 34.3, 0.344, 2.3, 633456.475, 3.0]),
                0.4,
            ),
            (
                np.asarray([4.45, 4.434, 8.45, 0.342, 7.34, 6346.3475, 3.0]),
                np.asarray([4.45, 4.434, 8.45, 0.342, 7.34, 6346.3475, 3.0]),
                0.1,
            ),
            (
                np.asarray([4, 65, 2, 677, 7, 6]),
                np.asarray([7, 7, 7, 7, 7, 7]),
                0.9,
            ),
        ),
    )
    def test_pinball_loss_scorer(
        y_true: np.ndarray, y_pred: np.ndarray, alpha: float
    ) -> None:
        """Test `pinball_loss_scorer`."""
        actual = pinball_loss_scorer(0.1)(y_true, y_pred)
        expected = mean_pinball_loss(y_true, y_pred, alpha=0.1)

        assert actual == expected

    @staticmethod
    @pytest.mark.parametrize("alpha", (None, 0.3))
    def test_optimize_single_pool(
        monkeypatch: MonkeyPatch,
        observations: np.ndarray,
        no_action: Callable[[Any], None],
        alpha: float,
    ) -> None:
        """Test `optimize_single_pool`."""
        monkeypatch.setattr(optuna.Study, "optimize", no_action)
        res = optimize_single_pool(observations, 0, alpha=alpha)
        assert isinstance(res, optuna.Study)

    @staticmethod
    @pytest.mark.parametrize("safely", (True, False))
    def test_get_best_params(study: MagicMock, safely: bool) -> None:
        """Test `get_best_params`."""
        if safely:
            res = get_best_params(study, safely)
            assert res == ({"x": 2.0}, False)
        else:
            with pytest.raises(ValueError, match="Study has no trials finished!"):
                get_best_params(study, safely)

    @staticmethod
    def test_optimize(
        monkeypatch: MonkeyPatch,
        observations: np.ndarray,
        no_action: Callable[[Any], None],
    ) -> None:
        """Test `optimize`."""
        path_to_module = "packages.valory.skills.apy_estimation_abci.ml.optimization"
        monkeypatch.setattr(
            f"{path_to_module}.optimize_single_pool", lambda *_, **__: MagicMock()
        )
        monkeypatch.setattr(
            f"{path_to_module}.get_best_params", lambda _: {"test_param": 0.23}
        )
        assert optimize({"test1": pd.DataFrame()}) == {"test1": {"test_param": 0.23}}


class TestObjective:
    """Tests for `Objective`."""

    @staticmethod
    def test__call__(monkeypatch: MonkeyPatch, observations: np.ndarray) -> None:
        """Test `__call__`."""
        # Make the sampler behave in a deterministic way.
        sampler = optuna.samplers.TPESampler(seed=0)
        # Create a study.
        study = optuna.create_study(sampler=sampler)

        # Create the Objective function.
        monkeypatch.setattr(
            pmdarima.model_selection,
            "cross_val_score",
            lambda *_, **__: [1, 2, 3, 4, 5],
        )
        objective_func = Objective(observations, "smape")

        # Start the optimization of the study.
        study.optimize(objective_func, n_trials=1)
        assert study.best_value == 3
        assert study.best_params == {"p": 3, "q": 3, "d": 2, "m": 15, "k": 3}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Dialogues classes for the ABCI connection."""

from aea.protocols.base import Address, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue

from packages.valory.connections.abci import PUBLIC_ID
from packages.valory.protocols.abci.dialogues import AbciDialogue as BaseAbciDialogue
from packages.valory.protocols.abci.dialogues import AbciDialogues as BaseAbciDialogues


AbciDialogue = BaseAbciDialogue


class AbciDialogues(BaseAbciDialogues):
    """The dialogues class keeps track of all ABCI dialogues."""

    def __init__(self) -> None:
        """Initialize dialogues."""

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return AbciDialogue.Role.CLIENT

        BaseAbciDialogues.__init__(
            self,
            self_address=str(PUBLIC_ID),
            role_from_first_message=role_from_first_message,
            dialogue_class=AbciDialogue,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Custom objects for the APY estimation ABCI application."""

from typing import Any

from packages.valory.skills.abstract_round_abci.models import ApiSpecs, BaseParams
from packages.valory.skills.abstract_round_abci.models import (
    BenchmarkTool as BaseBenchmarkTool,
)
from packages.valory.skills.abstract_round_abci.models import Requests as BaseRequests
from packages.valory.skills.abstract_round_abci.models import (
    SharedState as BaseSharedState,
)
from packages.valory.skills.apy_estimation_abci.rounds import APYEstimationAbciApp


Requests = BaseRequests
BenchmarkTool = BaseBenchmarkTool


class SharedState(BaseSharedState):
    """Keep the current shared state of the skill."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the state."""
        super().__init__(*args, abci_app_cls=APYEstimationAbciApp, **kwargs)


class RandomnessApi(ApiSpecs):
    """A model that wraps ApiSpecs for randomness api specifications."""


class FantomSubgraph(ApiSpecs):
    """A model that wraps ApiSpecs for Fantom subgraph specifications."""


class SpookySwapSubgraph(ApiSpecs):
    """A model that wraps ApiSpecs for SpookySwap subgraph specifications."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize SpookySwapSubgraph."""
        self.bundle_id: int = self.ensure("bundle_id", kwargs)
        super().__init__(*args, **kwargs)


class APYParams(BaseParams):  # pylint: disable=too-many-instance-attributes
    """Parameters."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the parameters object."""
        self.history_duration = self._ensure("history_duration", kwargs)
        self.optimizer_params = self._ensure("optimizer", kwargs)
        self.testing = self._ensure("testing", kwargs)
        self.estimation = self._ensure("estimation", kwargs)
        self.pair_ids = self._ensure("pair_ids", kwargs)
        self.ipfs_domain_name = self._ensure("ipfs_domain_name", kwargs)
        super().__init__(*args, **kwargs)

        self.__validate_params()

    def __validate_params(self) -> None:
        """Validate the given parameters."""
        # Eventually, we should probably validate all the parameters.
        if self.optimizer_params.get("timeout") is None:
            self.optimizer_params["timeout"] = None
        elif not isinstance(self.optimizer_params["timeout"], int):
            raise ValueError(
                "Parameter `timeout` can be either of type `int` or `None`. "
                f"{self.optimizer_params['timeout']} was given."
            )

        if self.optimizer_params.get("window_size") is None:
            self.optimizer_params["window_size"] = None
        elif not isinstance(self.optimizer_params["window_size"], int):
            raise ValueError(  # pragma: nocover
                "Parameter `window_size` can be either of type `int` or `None`. "
                f"{self.optimizer_params['window_size']} was given."
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/apy_estimation_abci skill's behaviours."""
import binascii
import json
import logging
import os
import time
from datetime import datetime
from enum import Enum
from itertools import product
from multiprocessing.pool import AsyncResult
from pathlib import Path, PosixPath
from typing import Any, Callable, Dict, Tuple, Type, Union, cast
from unittest import mock
from uuid import uuid4

import numpy as np
import optuna
import pandas as pd
import pytest
from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from aea.skills.tasks import TaskManager

from packages.valory.protocols.abci import AbciMessage  # noqa: F401
from packages.valory.skills.abstract_round_abci.base import AbciApp, AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    IPFSBehaviour,
)
from packages.valory.skills.abstract_round_abci.io.store import SupportedFiletype
from packages.valory.skills.abstract_round_abci.models import ApiSpecs, BenchmarkTool
from packages.valory.skills.apy_estimation_abci.behaviours import (
    APYEstimationBaseBehaviour,
    CycleResetBehaviour,
    EstimateBehaviour,
    EstimatorRoundBehaviour,
    FetchBatchBehaviour,
    FetchBehaviour,
    FreshModelResetBehaviour,
    OptimizeBehaviour,
    PrepareBatchBehaviour,
    PreprocessBehaviour,
    RandomnessBehaviour,
)
from packages.valory.skills.apy_estimation_abci.behaviours import (
    TestBehaviour as _TestBehaviour,
)
from packages.valory.skills.apy_estimation_abci.behaviours import (
    TrainBehaviour,
    TransformBehaviour,
    UpdateForecasterBehaviour,
)
from packages.valory.skills.apy_estimation_abci.ml.forecasting import (
    PoolIdToForecasterType,
    PoolIdToTestReportType,
)
from packages.valory.skills.apy_estimation_abci.ml.optimization import (
    PoolToHyperParamsWithStatusType,
)
from packages.valory.skills.apy_estimation_abci.ml.preprocessing import (
    prepare_pair_data,
)
from packages.valory.skills.apy_estimation_abci.rounds import Event, SynchronizedData
from packages.valory.skills.apy_estimation_abci.tools.etl import ResponseItemType

from tests.conftest import ROOT_DIR
from tests.test_skills.base import FSMBehaviourBaseCase
from tests.test_skills.test_apy_estimation_abci.conftest import DummyPipeline


SLEEP_TIME_TWEAK = 0.01

ipfs_daemon = pytest.mark.usefixtures("ipfs_daemon")


class DummyAsyncResult(object):
    """Dummy class for AsyncResult."""

    def __init__(
        self,
        task_result: Any,
        ready: bool = True,
    ) -> None:
        """Initialize class."""

        self.id = uuid4()
        self._ready = ready
        self._task_result = task_result

    def ready(
        self,
    ) -> bool:
        """Returns bool"""
        return self._ready

    def get(
        self,
    ) -> Any:
        """Returns task result."""
        return self._task_result


@ipfs_daemon
class APYEstimationFSMBehaviourBaseCase(FSMBehaviourBaseCase):
    """Base case for testing APYEstimation FSMBehaviour."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "valory", "skills", "apy_estimation_abci"
    )

    behaviour: EstimatorRoundBehaviour
    behaviour_class: Type[APYEstimationBaseBehaviour]
    next_behaviour_class: Type[APYEstimationBaseBehaviour]
    synchronized_data: SynchronizedData

    @classmethod
    def setup(cls, **kwargs: Any) -> None:
        """Set up the test class."""
        super().setup(
            param_overrides={"ipfs_domain_name": "/dns/localhost/tcp/5001/http"}
        )
        cls.synchronized_data = SynchronizedData(
            AbciAppDB(
                setup_data={"full_training": [False]},
            )
        )

    def end_round(self, done_event: Enum = Event.DONE) -> None:
        """Ends round early to cover `wait_for_end` generator."""
        super().end_round(done_event)


class TestFetchAndBatchBehaviours(APYEstimationFSMBehaviourBaseCase):
    """Test FetchBehaviour and FetchBatchBehaviour."""

    behaviour_class = FetchBehaviour
    next_behaviour_class = TransformBehaviour

    @pytest.mark.parametrize("batch_flag", (True, False))
    def test_setup(self, monkeypatch: MonkeyPatch, batch_flag: bool) -> None:
        """Test behaviour setup."""
        self.skill.skill_context.state.round_sequence.abci_app._last_timestamp = (
            datetime.now()
        )

        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            self.synchronized_data,
        )
        cast(FetchBehaviour, self.behaviour.current_behaviour).batch = batch_flag

        monkeypatch.setattr(os.path, "join", lambda *_: "")
        cast(APYEstimationBaseBehaviour, self.behaviour.current_behaviour).setup()

    def test_handle_response(self, caplog: LogCaptureFixture) -> None:
        """Test `handle_response`."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            self.synchronized_data,
        )
        cast(
            FetchBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK

        # test with empty response.
        specs = ApiSpecs(
            url="test",
            api_id="test",
            method="GET",
            name="test",
            skill_context=self.behaviour.context,
        )

        with caplog.at_level(
            logging.ERROR,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            handling_generator = cast(
                FetchBehaviour, self.behaviour.current_behaviour
            )._handle_response(None, "test_context", ("", 0), specs)
            next(handling_generator)
            time.sleep(SLEEP_TIME_TWEAK + 0.01)

            try:
                next(handling_generator)
            except StopIteration as res:
                assert res.value is None

            assert (
                "[test_agent_name] Could not get test_context from test" in caplog.text
            )
            assert specs._retries_attempted == 1

        caplog.clear()
        with caplog.at_level(
            logging.INFO,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            handling_generator = cast(
                FetchBehaviour, self.behaviour.current_behaviour
            )._handle_response({"test": [4, 5]}, "test", ("test", 0), specs)
            try:
                next(handling_generator)
            except StopIteration as res:
                assert res.value == 4
            assert "[test_agent_name] Retrieved test: 4." in caplog.text
            assert specs._retries_attempted == 0

    def test_fetch_behaviour(
        self,
        block_from_timestamp_q: str,
        eth_price_usd_q: str,
        pairs_q: str,
        pool_fields: Tuple[str, ...],
    ) -> None:
        """Run tests."""
        history_duration = cast(
            FetchBehaviour, self.behaviour.current_behaviour
        ).params.history_duration
        self.skill.skill_context.state.round_sequence.abci_app._last_timestamp = (
            datetime.utcfromtimestamp(1618735147 + history_duration * 30 * 24 * 60 * 60)
        )

        self.fast_forward_to_behaviour(
            self.behaviour, FetchBehaviour.behaviour_id, self.synchronized_data
        )
        cast(FetchBehaviour, self.behaviour.current_behaviour).params.pair_ids = [
            "0xec454eda10accdd66209c57af8c12924556f3abd"
        ]

        request_kwargs: Dict[str, Union[str, bytes]] = dict(
            method="POST",
            url="https://api.thegraph.com/subgraphs/name/eerieeight/spookyswap",
            headers="Content-Type: application/json\r\n",
            version="",
        )
        response_kwargs = dict(
            version="",
            status_code=200,
            status_text="",
            headers="",
        )

        # block request.
        request_kwargs[
            "url"
        ] = "https://api.thegraph.com/subgraphs/name/matthewlilley/fantom-blocks"
        request_kwargs["body"] = json.dumps({"query": block_from_timestamp_q}).encode(
            "utf-8"
        )
        res = {"data": {"blocks": [{"timestamp": "1", "number": "3830367"}]}}
        response_kwargs["body"] = json.dumps(res).encode("utf-8")
        self.behaviour.act_wrapper()
        self.mock_http_request(request_kwargs, response_kwargs)

        # ETH price request.
        request_kwargs[
            "url"
        ] = "https://api.thegraph.com/subgraphs/name/eerieeight/spookyswap"
        request_kwargs["body"] = json.dumps({"query": eth_price_usd_q}).encode("utf-8")
        res = {"data": {"bundles": [{"ethPrice": "0.8973548"}]}}
        response_kwargs["body"] = json.dumps(res).encode("utf-8")
        self.behaviour.act_wrapper()
        self.mock_http_request(request_kwargs, response_kwargs)

        # top pairs data.
        request_kwargs["body"] = json.dumps({"query": pairs_q}).encode("utf-8")
        res = {
            "data": {
                "pairs": [
                    {field: dummy_value for field in pool_fields}
                    for dummy_value in ("dum1", "dum2")
                ]
            }
        }
        response_kwargs["body"] = json.dumps(res).encode("utf-8")
        self.behaviour.act_wrapper()
        self.mock_http_request(request_kwargs, response_kwargs)

        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == TransformBehaviour.behaviour_id

    def test_fetch_behaviour_retries_exceeded(
        self, monkeypatch: MonkeyPatch, caplog: LogCaptureFixture
    ) -> None:
        """Run tests for exceeded retries."""
        self.skill.skill_context.state.round_sequence.abci_app._last_timestamp = (
            datetime.now()
        )

        self.fast_forward_to_behaviour(
            self.behaviour, FetchBehaviour.behaviour_id, self.synchronized_data
        )

        subgraphs_sorted_by_utilization_moment: Tuple[Any, ...] = (
            self.behaviour.context.spooky_subgraph,
            self.behaviour.context.fantom_subgraph,
            self.behaviour.context.spooky_subgraph,
            self.behaviour.context.spooky_subgraph,
        )

        for subgraph in subgraphs_sorted_by_utilization_moment:
            monkeypatch.setattr(subgraph, "is_retries_exceeded", lambda *_: True)
            with caplog.at_level(
                logging.ERROR,
                logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
            ):
                self.behaviour.act_wrapper()
                self.behaviour.act_wrapper()
            assert (
                "Retries were exceeded while downloading the historical data!"
                in caplog.text
            )

    def test_fetch_value_none(
        self,
        caplog: LogCaptureFixture,
        block_from_timestamp_q: str,
        eth_price_usd_q: str,
        pairs_q: str,
        pool_fields: Tuple[str, ...],
    ) -> None:
        """Test when fetched value is none."""
        history_duration = cast(
            FetchBehaviour, self.behaviour.current_behaviour
        ).params.history_duration
        self.skill.skill_context.state.round_sequence.abci_app._last_timestamp = (
            datetime.utcfromtimestamp(1618735147 + history_duration * 30 * 24 * 60 * 60)
        )
        self.fast_forward_to_behaviour(
            self.behaviour, FetchBehaviour.behaviour_id, self.synchronized_data
        )
        cast(FetchBehaviour, self.behaviour.current_behaviour).params.pair_ids = [
            "0xec454eda10accdd66209c57af8c12924556f3abd"
        ]
        cast(
            FetchBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK

        request_kwargs: Dict[str, Union[str, bytes]] = dict(
            method="POST",
            url="https://api.thegraph.com/subgraphs/name/eerieeight/spookyswap",
            headers="Content-Type: application/json\r\n",
            version="",
            body=b"",
        )
        response_kwargs = dict(
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=b"",
        )

        # block request with None response.
        request_kwargs[
            "url"
        ] = "https://api.thegraph.com/subgraphs/name/matthewlilley/fantom-blocks"
        request_kwargs["body"] = json.dumps({"query": block_from_timestamp_q}).encode(
            "utf-8"
        )
        response_kwargs["body"] = b""

        with caplog.at_level(
            logging.ERROR,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.act_wrapper()
            self.mock_http_request(request_kwargs, response_kwargs)
        assert "[test_agent_name] Could not get block from fantom" in caplog.text

        caplog.clear()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        # block request.
        request_kwargs[
            "url"
        ] = "https://api.thegraph.com/subgraphs/name/matthewlilley/fantom-blocks"
        request_kwargs["body"] = json.dumps({"query": block_from_timestamp_q}).encode(
            "utf-8"
        )
        res = {"data": {"blocks": [{"timestamp": "1", "number": "3830367"}]}}
        response_kwargs["body"] = json.dumps(res).encode("utf-8")
        self.behaviour.act_wrapper()
        self.mock_http_request(request_kwargs, response_kwargs)

        # ETH price request with None response.
        request_kwargs[
            "url"
        ] = "https://api.thegraph.com/subgraphs/name/eerieeight/spookyswap"
        request_kwargs["body"] = json.dumps({"query": eth_price_usd_q}).encode("utf-8")
        response_kwargs["body"] = b""

        with caplog.at_level(
            logging.ERROR,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.act_wrapper()
            self.mock_http_request(request_kwargs, response_kwargs)
        assert (
            "[test_agent_name] Could not get ETH price for block "
            "{'timestamp': '1', 'number': '3830367'} from spookyswap" in caplog.text
        )

        caplog.clear()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        # block request.
        request_kwargs[
            "url"
        ] = "https://api.thegraph.com/subgraphs/name/matthewlilley/fantom-blocks"
        request_kwargs["body"] = json.dumps({"query": block_from_timestamp_q}).encode(
            "utf-8"
        )
        res = {"data": {"blocks": [{"timestamp": "1", "number": "3830367"}]}}
        response_kwargs["body"] = json.dumps(res).encode("utf-8")
        self.behaviour.act_wrapper()
        self.mock_http_request(request_kwargs, response_kwargs)

        # ETH price request.
        request_kwargs[
            "url"
        ] = "https://api.thegraph.com/subgraphs/name/eerieeight/spookyswap"
        request_kwargs["body"] = json.dumps({"query": eth_price_usd_q}).encode("utf-8")
        res = {"data": {"bundles": [{"ethPrice": "0.8973548"}]}}
        response_kwargs["body"] = json.dumps(res).encode("utf-8")
        self.behaviour.act_wrapper()
        self.mock_http_request(request_kwargs, response_kwargs)

        # top pairs data with None response.
        request_kwargs["body"] = json.dumps({"query": pairs_q}).encode("utf-8")
        response_kwargs["body"] = b""

        with caplog.at_level(
            logging.ERROR,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.act_wrapper()
            self.mock_http_request(request_kwargs, response_kwargs)
        assert (
            "[test_agent_name] Could not get pool data for block {'timestamp': '1', 'number': '3830367'} "
            "from spookyswap" in caplog.text
        )

        caplog.clear()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

    @pytest.mark.parametrize("total_days", (0, 3))
    def test_fetch_behaviour_stop_iteration(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        caplog: LogCaptureFixture,
        no_action: Callable[[Any], None],
        total_days: int,
    ) -> None:
        """Test `FetchBehaviour`'s `async_act` after all the timestamps have been generated."""
        self.skill.skill_context.state.round_sequence.abci_app._last_timestamp = (
            datetime.now()
        )

        # fast-forward to fetch behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour, FetchBehaviour.behaviour_id, self.synchronized_data
        )
        # set history duration to a negative value in order to raise a `StopIteration`.
        cast(
            FetchBehaviour, self.behaviour.current_behaviour
        ).params.history_duration = -1

        # test empty retrieved history.
        with caplog.at_level(
            logging.ERROR,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.act_wrapper()
        assert "Could not download any historical data!" in caplog.text
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        # fast-forward to fetch behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour, FetchBehaviour.behaviour_id, self.synchronized_data
        )

        # test with retrieved history and valid save path.
        current_behaviour = cast(FetchBehaviour, self.behaviour.current_behaviour)
        current_behaviour._pairs_hist = [{"pool1": "test"}, {"pool2": "test"}]
        current_behaviour._total_days = total_days
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == TransformBehaviour.behaviour_id

        # fast-forward to fetch behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour, FetchBehaviour.behaviour_id, self.synchronized_data
        )

    def test_clean_up(
        self,
    ) -> None:
        """Test clean-up."""
        self.fast_forward_to_behaviour(
            self.behaviour, FetchBehaviour.behaviour_id, self.synchronized_data
        )

        self.behaviour.context.spooky_subgraph._retries_attempted = 1
        self.behaviour.context.fantom_subgraph._retries_attempted = 1
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour.clean_up()
        assert self.behaviour.context.spooky_subgraph._retries_attempted == 0
        assert self.behaviour.context.fantom_subgraph._retries_attempted == 0


class TestTransformBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test TransformBehaviour."""

    behaviour_class = TransformBehaviour
    next_behaviour_class = PreprocessBehaviour

    def _fast_forward(self, tmp_path: PosixPath, ipfs_succeed: bool = True) -> None:
        """Setup `TestTransformBehaviour`."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Send historical data to IPFS and get the hash.
        if ipfs_succeed:
            hash_ = cast(BaseBehaviour, self.behaviour.current_behaviour).send_to_ipfs(
                os.path.join(
                    tmp_path,
                    f"historical_data_period_{self.synchronized_data.period_count}.json",
                ),
                {"test": "test"},
                filetype=SupportedFiletype.JSON,
            )
        else:
            hash_ = "test"

        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(
                AbciAppDB(
                    setup_data=dict(
                        most_voted_randomness=[0], most_voted_history=[hash_]
                    ),
                )
            ),
        )

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    def test_setup(self, tmp_path: PosixPath) -> None:
        """Test behaviour setup."""
        self._fast_forward(tmp_path)
        self.behaviour.context.task_manager.start()
        cast(TransformBehaviour, self.behaviour.current_behaviour).setup()

    def test_task_not_ready(
        self,
        monkeypatch: MonkeyPatch,
        caplog: LogCaptureFixture,
        tmp_path: PosixPath,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
    ) -> None:
        """Run test for `transform_behaviour` when task result is not ready."""
        self._fast_forward(tmp_path)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(
                transformed_historical_data_no_datetime_conversion, ready=False
            ),
        )

        with caplog.at_level(
            logging.DEBUG,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.context.task_manager.start()

            cast(
                TransformBehaviour, self.behaviour.current_behaviour
            ).params.sleep_time = SLEEP_TIME_TWEAK
            self.behaviour.act_wrapper()

            # Sleep to wait for the behaviour that is also sleeping.
            time.sleep(SLEEP_TIME_TWEAK + 0.01)

            # Continue the `async_act` after the sleep of the Behaviour.
            self.behaviour.act_wrapper()

        assert "[test_agent_name] Entered in the 'transform' behaviour" in caplog.text

        assert (
            "[test_agent_name] The transform task is not finished yet." in caplog.text
        )

        self.end_round()

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_transform_behaviour(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `transform_behaviour`."""
        self._fast_forward(tmp_path, ipfs_succeed)

        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.tasks.transform_hist_data",
            lambda _: transformed_historical_data_no_datetime_conversion,
        )
        monkeypatch.setattr(
            self._skill._skill_context._agent_context._task_manager,  # type: ignore
            "get_task_result",
            lambda *_: DummyAsyncResult(
                transformed_historical_data_no_datetime_conversion
            ),
        )
        monkeypatch.setattr(
            self._skill._skill_context._agent_context._task_manager,  # type: ignore
            "enqueue_task",
            lambda *_, **__: 3,
        )

        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestPreprocessBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test PreprocessBehaviour."""

    behaviour_class = PreprocessBehaviour
    next_behaviour_class = RandomnessBehaviour

    def _fast_forward(
        self,
        data_found: bool,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
    ) -> pd.DataFrame:
        """Fast-forward to behaviour."""
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore
        if data_found:
            monkeypatch.setattr(
                IPFSBehaviour, "get_from_ipfs", lambda *_, **__: {"test": "test"}
            )
        # Increase the amount of dummy data for the train-test split,
        # as many times as the threshold in `group_and_filter_pair_data`.
        transformed_historical_data = pd.DataFrame(
            np.repeat(
                transformed_historical_data_no_datetime_conversion.values, 5, axis=0
            ),
            columns=transformed_historical_data_no_datetime_conversion.columns,
        )

        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(AbciAppDB(setup_data=dict(most_voted_transform=["test"]))),
        )
        behaviour = cast(PreprocessBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.behaviour_class.behaviour_id

        behaviour.params.sleep_time = SLEEP_TIME_TWEAK

        return transformed_historical_data

    @pytest.mark.parametrize(
        "data_found, task_ready", ((True, True), (True, False), (False, False))
    )
    def test_preprocess_behaviour(
        self,
        data_found: bool,
        task_ready: bool,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
    ) -> None:
        """Run test for `preprocess_behaviour`."""
        transformed_historical_data = self._fast_forward(
            data_found,
            transformed_historical_data_no_datetime_conversion,
            monkeypatch,
            tmp_path,
        )

        # Convert the `blockTimestamp` to a pandas datetime.
        transformed_historical_data["blockTimestamp"] = pd.to_datetime(
            transformed_historical_data["blockTimestamp"], unit="s"
        )
        monkeypatch.setattr(
            self._skill._skill_context._agent_context._task_manager,  # type: ignore
            "get_task_result",
            lambda *_: DummyAsyncResult(
                prepare_pair_data(transformed_historical_data), task_ready
            ),
        )
        monkeypatch.setattr(
            self._skill._skill_context._agent_context._task_manager,  # type: ignore
            "enqueue_task",
            lambda *_, **__: 3,
        )

        self.behaviour.act_wrapper()

        if data_found:
            assert (
                cast(PreprocessBehaviour, self.behaviour.current_behaviour)._pairs_hist
                is not None
            ), "Pairs history could not be loaded!"

        if task_ready:
            self.mock_a2a_transaction()
            self._test_done_flag_set()
            self.end_round()
            behaviour = cast(PreprocessBehaviour, self.behaviour.current_behaviour)
            assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id

        else:
            self.behaviour.act_wrapper()
            time.sleep(SLEEP_TIME_TWEAK + 0.01)
            self.behaviour.act_wrapper()
            behaviour = cast(PreprocessBehaviour, self.behaviour.current_behaviour)
            assert behaviour.behaviour_id == self.behaviour_class.behaviour_id


class TestPrepareBatchBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test `PrepareBatchBehaviour`."""

    behaviour_class = PrepareBatchBehaviour
    next_behaviour_class = UpdateForecasterBehaviour

    def _fast_forward(
        self,
        tmp_path: PosixPath,
        transformed_historical_data: pd.DataFrame,
        batch: ResponseItemType,
        ipfs_succeed: bool = True,
    ) -> None:
        """Setup `PrepareBatchBehaviour`."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Create a dictionary with all the dummy data to send to IPFS.
        data_to_send = {
            "hist": {
                "filepath": os.path.join(
                    tmp_path,
                    f"latest_observations_period_{self.synchronized_data.period_count - 1}.csv",
                ),
                "obj": transformed_historical_data.iloc[[0, 2]].reset_index(drop=True),
                "filetype": SupportedFiletype.CSV,
            },
            "batch": {
                "filepath": os.path.join(
                    tmp_path,
                    f"historical_data_batch_0_period_{self.synchronized_data.period_count}.json",
                ),
                "obj": batch,
                "filetype": SupportedFiletype.JSON,
            },
        }

        # Send dummy data to IPFS and get the hashes.
        if ipfs_succeed:
            hashes = {}
            for item_name, item_args in data_to_send.items():
                hashes[item_name] = cast(
                    BaseBehaviour, self.behaviour.current_behaviour
                ).send_to_ipfs(**item_args)
        else:
            hashes = {item_name: "test" for item_name, _ in data_to_send.items()}

        # fast-forward to the `PrepareBatchBehaviour` behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            latest_observation_hist_hash=hashes["hist"],
                            most_voted_batch=hashes["batch"],
                            latest_observation_timestamp=0,
                        )
                    ),
                )
            ),
        )

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    def test_prepare_batch_behaviour_setup(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
        batch: ResponseItemType,
        prepare_batch_task_result: Dict[str, pd.DataFrame],
    ) -> None:
        """Test behaviour setup."""
        self._fast_forward(
            tmp_path, transformed_historical_data_no_datetime_conversion, batch
        )

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(prepare_batch_task_result),
        )

        current_behaviour = cast(
            PrepareBatchBehaviour, self.behaviour.current_behaviour
        )
        current_behaviour.setup()
        assert not any(batch is None for batch in current_behaviour._batches)

    def test_task_not_ready(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
        batch: ResponseItemType,
        prepare_batch_task_result: Dict[str, pd.DataFrame],
    ) -> None:
        """Run test for behaviour when task result is not ready."""
        self._fast_forward(
            tmp_path,
            transformed_historical_data_no_datetime_conversion,
            batch,
        )

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(prepare_batch_task_result, ready=False),
        )

        cast(
            PrepareBatchBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK
        self.behaviour.act_wrapper()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_prepare_batch_behaviour(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
        batch: ResponseItemType,
        prepare_batch_task_result: Dict[str, pd.DataFrame],
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `prepare_behaviour`."""
        self._fast_forward(
            tmp_path,
            transformed_historical_data_no_datetime_conversion,
            batch,
            ipfs_succeed,
        )

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(prepare_batch_task_result),
        )

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.behaviour_class.behaviour_id

        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestRandomnessBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test RandomnessBehaviour."""

    randomness_behaviour_class = RandomnessBehaviour  # type: ignore
    next_behaviour_class = OptimizeBehaviour

    drand_response = {
        "round": 1416669,
        "randomness": "f6be4bf1fa229f22340c1a5b258f809ac4af558200775a67dacb05f0cb258a11",
        "signature": (
            "b44d00516f46da3a503f9559a634869b6dc2e5d839e46ec61a090e3032172954929a5"
            "d9bd7197d7739fe55db770543c71182562bd0ad20922eb4fe6b8a1062ed21df3b68de"
            "44694eb4f20b35262fa9d63aa80ad3f6172dd4d33a663f21179604"
        ),
        "previous_signature": (
            "903c60a4b937a804001032499a855025573040cb86017c38e2b1c3725286756ce8f33"
            "61188789c17336beaf3f9dbf84b0ad3c86add187987a9a0685bc5a303e37b008fba8c"
            "44f02a416480dd117a3ff8b8075b1b7362c58af195573623187463"
        ),
    }

    def test_randomness_behaviour(
        self,
    ) -> None:
        """Test RandomnessBehaviour."""

        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps(self.drand_response).encode("utf-8"),
            ),
        )

        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id

    def test_invalid_drand_value(
        self,
    ) -> None:
        """Test invalid drand values."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()

        drand_invalid = self.drand_response.copy()
        drand_invalid["randomness"] = binascii.hexlify(b"randomness_hex").decode()
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps(drand_invalid).encode(),
            ),
        )

    def test_invalid_response(
        self,
    ) -> None:
        """Test invalid json response."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        cast(
            RandomnessBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK

        self.behaviour.act_wrapper()

        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="", status_code=200, status_text="", headers="", body=b""
            ),
        )
        self.behaviour.act_wrapper()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

    def test_max_retries_reached(
        self,
    ) -> None:
        """Test with max retries reached."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        with mock.patch.object(
            self.behaviour.context.randomness_api,
            "is_retries_exceeded",
            return_value=True,
        ):
            self.behaviour.act_wrapper()
            behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
            assert (
                behaviour.behaviour_id == self.randomness_behaviour_class.behaviour_id
            )
            self._test_done_flag_set()

    def test_clean_up(
        self,
    ) -> None:
        """Test when `observed` value is none."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        self.behaviour.context.randomness_api._retries_attempted = 1
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour.clean_up()
        assert self.behaviour.context.randomness_api._retries_attempted == 0


class TestOptimizeBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test OptimizeBehaviour."""

    behaviour_class = OptimizeBehaviour
    next_behaviour_class = TrainBehaviour

    def _fast_forward(
        self,
        tmp_path: PosixPath,
        ipfs_succeed: bool = True,
    ) -> None:
        """Setup `OptimizeBehaviour`."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Create a dictionary with all the dummy data to send to IPFS.
        data_to_send = {}
        for split in ("train", "test"):
            data_to_send[split] = {
                "filepath": os.path.join(
                    tmp_path,
                    f"y_{split}",
                    f"period_{self.synchronized_data.period_count}",
                ),
                "obj": {
                    f"{split}_{i}": pd.DataFrame([i for i in range(5)])
                    for i in range(3)
                },
                "multiple": True,
                "filetype": SupportedFiletype.CSV,
            }

        # Send dummy data to IPFS and get the hashes.
        if ipfs_succeed:
            hashes = {}
            for item_name, item_args in data_to_send.items():
                hashes[item_name] = cast(
                    BaseBehaviour, self.behaviour.current_behaviour
                ).send_to_ipfs(**item_args)
        else:
            hashes = {item_name: "non_existing" for item_name in data_to_send.keys()}

        # fast-forward to the `OptimizeBehaviour` behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(
                AbciAppDB(
                    setup_data=dict(
                        most_voted_randomness=[0],
                        most_voted_split=[hashes["train"] + hashes["test"]],
                    ),
                )
            ),
        )

        assert (
            cast(OptimizeBehaviour, self.behaviour.current_behaviour).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    def test_setup(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        optimize_task_result_empty: optuna.Study,
    ) -> None:
        """Test behaviour setup."""
        self._fast_forward(tmp_path)

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(optimize_task_result_empty),
        )
        current_behaviour = cast(OptimizeBehaviour, self.behaviour.current_behaviour)
        current_behaviour.setup()
        assert current_behaviour._y is not None

    def test_task_not_ready(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        optimize_task_result_empty: optuna.Study,
    ) -> None:
        """Run test for behaviour when task result is not ready."""
        self._fast_forward(tmp_path)

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(optimize_task_result_empty, ready=False),
        )

        cast(
            OptimizeBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK
        self.behaviour.act_wrapper()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_optimize_behaviour(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        caplog: LogCaptureFixture,
        optimize_task_result: PoolToHyperParamsWithStatusType,
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `optimize_behaviour`."""
        self._fast_forward(tmp_path, ipfs_succeed)

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 3)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(optimize_task_result),
        )

        with caplog.at_level(
            logging.WARNING,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            # note: this will capture the warning only if the first item in the `optimize_task_result` dict
            # is the one representing the non-finished trial, i.e., `False`.
            self.behaviour.act_wrapper()

        if ipfs_succeed:
            for _ in range(len(optimize_task_result)):
                self.behaviour.act_wrapper()
            assert (
                "The optimization could not be done for pool `test1`! "
                "Please make sure that there is a sufficient number of data for the optimization procedure. "
                "Parameters have been set randomly!"
            ) in caplog.text

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestTrainBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test TrainBehaviour."""

    behaviour_class = TrainBehaviour
    next_behaviour_class = _TestBehaviour

    def _fast_forward(
        self,
        tmp_path: PosixPath,
        ipfs_succeed: bool = True,
        full_training: bool = False,
    ) -> None:
        """Setup `TestTrainBehaviour`."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Create a dictionary with all the dummy data to send to IPFS.
        data_to_send = {
            "params": {
                "filepath": os.path.join(
                    tmp_path,
                    "best_params",
                    f"period_{self.synchronized_data.period_count}",
                ),
                "obj": {
                    "pool1.json": {"p": 1, "q": 1, "d": 1, "m": 1},
                    "pool2.json": {"p": 2, "q": 2, "d": 1, "m": 1},
                },
                "multiple": True,
                "filetype": SupportedFiletype.JSON,
            }
        }
        for split in ("train", "test"):
            data_to_send[split] = {
                "filepath": os.path.join(
                    tmp_path,
                    f"y_{split}",
                    f"period_{self.synchronized_data.period_count}",
                ),
                "obj": {
                    f"pool{i}.csv": pd.DataFrame([i for i in range(5)])
                    for i in range(3)
                },
                "multiple": True,
                "filetype": SupportedFiletype.CSV,
            }

        # Send dummy data to IPFS and get the hashes.
        if ipfs_succeed:
            hashes = {}
            for item_name, item_args in data_to_send.items():
                hashes[item_name] = cast(
                    BaseBehaviour, self.behaviour.current_behaviour
                ).send_to_ipfs(**item_args)
        else:
            hashes = {item_name: "non_existing" for item_name in data_to_send.keys()}

        # fast-forward to the `TrainBehaviour` behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            full_training=full_training,
                            most_voted_params=hashes["params"],
                            most_voted_split=hashes["train"] + hashes["test"],
                        )
                    ),
                )
            ),
        )

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize(
        "full_training, ipfs_succeed", product((True, False), repeat=2)
    )
    def test_setup(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        no_action: Callable[[Any], None],
        ipfs_succeed: bool,
        full_training: bool,
    ) -> None:
        """Test behaviour setup."""
        self._fast_forward(tmp_path, ipfs_succeed, full_training)

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(TaskManager, "get_task_result", no_action)

        current_behaviour = cast(TrainBehaviour, self.behaviour.current_behaviour)
        current_behaviour.setup()
        if ipfs_succeed:
            assert not any(
                arg is None
                for arg in (current_behaviour._y, current_behaviour._best_params)
            )
        else:
            assert all(
                arg is None
                for arg in (current_behaviour._y, current_behaviour._best_params)
            )

    def test_task_not_ready(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
    ) -> None:
        """Run test for behaviour when task result is not ready."""
        self._fast_forward(tmp_path)

        self.behaviour.context.task_manager.start()

        monkeypatch.setattr(AsyncResult, "ready", lambda *_: False)
        cast(
            TrainBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK
        self.behaviour.act_wrapper()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_train_behaviour(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        train_task_result: PoolIdToForecasterType,
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `train_behaviour`."""
        self._fast_forward(tmp_path, ipfs_succeed)

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 3)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(train_task_result),
        )

        # act.
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestTestBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test TestBehaviour."""

    behaviour_class = _TestBehaviour
    next_behaviour_class = TrainBehaviour

    def _fast_forward(
        self,
        tmp_path: PosixPath,
        ipfs_succeed: bool = True,
    ) -> None:
        """Setup `TestTrainBehaviour`."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Create a dictionary with all the dummy data to send to IPFS.
        data_to_send = {
            "model": {
                "filepath": os.path.join(
                    tmp_path,
                    "forecasters",
                    f"period_{self.synchronized_data.period_count}",
                ),
                "obj": {f"pool{i}.joblib": DummyPipeline() for i in range(3)},
                "multiple": True,
                "filetype": SupportedFiletype.PM_PIPELINE,
            }
        }
        for split in ("train", "test"):
            data_to_send[split] = {
                "filepath": os.path.join(
                    tmp_path,
                    f"y_{split}",
                    f"period_{self.synchronized_data.period_count}",
                ),
                "obj": {
                    f"pool{i}.csv": pd.DataFrame([i for i in range(5)])
                    for i in range(3)
                },
                "multiple": True,
                "filetype": SupportedFiletype.CSV,
            }

        # Send dummy data to IPFS and get the hashes.
        if ipfs_succeed:
            hashes = {}
            for item_name, item_args in data_to_send.items():
                hashes[item_name] = cast(
                    BaseBehaviour, self.behaviour.current_behaviour
                ).send_to_ipfs(**item_args)
        else:
            hashes = {item_name: "non_existing" for item_name in data_to_send.keys()}

        # fast-forward to the `TestBehaviour` behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(
                AbciAppDB(
                    setup_data=dict(
                        most_voted_models=[hashes["model"]],
                        most_voted_split=[hashes["train"] + hashes["test"]],
                    ),
                )
            ),
        )

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_setup(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        ipfs_succeed: bool,
        no_action: Callable[[Any], None],
    ) -> None:
        """Test behaviour setup."""
        self._fast_forward(tmp_path, ipfs_succeed)

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(TaskManager, "get_task_result", no_action)

        current_behaviour = cast(_TestBehaviour, self.behaviour.current_behaviour)
        current_behaviour.setup()

        is_none = (
            arg is None
            for arg in (
                getattr(current_behaviour, arg_name)
                for arg_name in ("_y_train", "_y_test", "_forecasters")
            )
        )
        if ipfs_succeed:
            assert not any(is_none)
        else:
            assert all(is_none)

    def test_task_not_ready(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
    ) -> None:
        """Run test for behaviour when task result is not ready."""
        self._fast_forward(tmp_path)

        self.behaviour.context.task_manager.start()
        monkeypatch.setattr(AsyncResult, "ready", lambda *_: False)
        cast(
            _TestBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK
        self.behaviour.act_wrapper()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_test_behaviour(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        ipfs_succeed: bool,
        _test_task_result: PoolIdToTestReportType,
    ) -> None:
        """Run test for `test_behaviour`."""
        self._fast_forward(tmp_path, ipfs_succeed)

        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 3)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(_test_task_result),
        )

        # test act.
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestUpdateForecasterBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test `UpdateForecasterBehaviour`."""

    behaviour_class = UpdateForecasterBehaviour
    next_behaviour_class = EstimateBehaviour

    def _fast_forward(
        self,
        tmp_path: PosixPath,
        prepare_batch_task_result: pd.DataFrame,
        ipfs_succeed: bool = True,
    ) -> None:
        """Setup `TestUpdateForecasterBehaviour`."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Create a dictionary with all the dummy data to send to IPFS.
        data_to_send = {
            "model": {
                "filepath": os.path.join(
                    tmp_path,
                    "fully_trained_forecasters",
                    f"period_{self.synchronized_data.period_count - 1}",
                ),
                "obj": {f"pool{i}.joblib": DummyPipeline() for i in range(3)},
                "multiple": True,
                "filetype": SupportedFiletype.PM_PIPELINE,
            },
            "observation": {
                "filepath": os.path.join(
                    tmp_path,
                    f"latest_observations_period_{self.synchronized_data.period_count}.csv",
                ),
                "obj": prepare_batch_task_result,
                "filetype": SupportedFiletype.CSV,
            },
        }

        # Send dummy data to IPFS and get the hashes.
        if ipfs_succeed:
            hashes = {}
            for item_name, item_args in data_to_send.items():
                hashes[item_name] = cast(
                    BaseBehaviour, self.behaviour.current_behaviour
                ).send_to_ipfs(**item_args)
        else:
            hashes = {
                item_name: "non_existing" for item_name, _ in data_to_send.items()
            }

        # fast-forward to the `TestBehaviour` behaviour.
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(
                AbciAppDB(
                    setup_data=dict(
                        most_voted_models=[hashes["model"]],
                        latest_observation_hist_hash=[hashes["observation"]],
                    ),
                )
            ),
        )

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_update_forecaster_setup(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        no_action: Callable[[Any], None],
        prepare_batch_task_result: pd.DataFrame,
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `UpdateForecasterBehaviour`'s setup method."""
        self._fast_forward(tmp_path, prepare_batch_task_result, ipfs_succeed)
        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(TaskManager, "get_task_result", no_action)

        current_behaviour = cast(
            UpdateForecasterBehaviour, self.behaviour.current_behaviour
        )
        current_behaviour.setup()

        is_none = (
            arg is None
            for arg in (
                getattr(current_behaviour, arg_name)
                for arg_name in ("_y", "_forecasters")
            )
        )
        if ipfs_succeed:
            assert not any(is_none)
        else:
            assert all(is_none)

    def test_task_not_ready(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        prepare_batch_task_result: pd.DataFrame,
    ) -> None:
        """Run test for behaviour when task result is not ready."""
        self._fast_forward(tmp_path, prepare_batch_task_result)

        self.behaviour.context.task_manager.start()

        monkeypatch.setattr(AsyncResult, "ready", lambda *_: False)
        cast(
            TrainBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK
        self.behaviour.act_wrapper()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_update_forecaster_behaviour(
        self,
        tmp_path: PosixPath,
        monkeypatch: MonkeyPatch,
        prepare_batch_task_result: pd.DataFrame,
        train_task_result: PoolIdToForecasterType,
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `UpdateForecasterBehaviour`."""
        self._fast_forward(tmp_path, prepare_batch_task_result, ipfs_succeed)
        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 3)
        monkeypatch.setattr(
            TaskManager,
            "get_task_result",
            lambda *_: DummyAsyncResult(train_task_result),
        )

        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()
        behaviour = cast(UpdateForecasterBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestEstimateBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test EstimateBehaviour."""

    behaviour_class = EstimateBehaviour
    next_behaviour_class = FreshModelResetBehaviour

    def _fast_forward(self, tmp_path: PosixPath, ipfs_succeed: bool = True) -> None:
        """Setup `TestTransformBehaviour`."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Send dummy forecasters to IPFS and get the hash.
        if ipfs_succeed:
            hash_ = cast(BaseBehaviour, self.behaviour.current_behaviour).send_to_ipfs(
                os.path.join(
                    tmp_path,
                    "fully_trained_forecasters",
                    f"period_{self.synchronized_data.period_count}",
                ),
                {f"pool{i}.joblib": DummyPipeline() for i in range(3)},
                multiple=True,
                filetype=SupportedFiletype.PM_PIPELINE,
            )
        else:
            hash_ = "non_existing"

        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            SynchronizedData(
                AbciAppDB(
                    setup_data=dict(most_voted_models=[hash_]),
                )
            ),
        )

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_estimate_setup(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        no_action: Callable[[Any], None],
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `EstimateBehaviour`'s setup method."""
        self._fast_forward(tmp_path, ipfs_succeed)
        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(TaskManager, "get_task_result", no_action)

        current_behaviour = cast(
            UpdateForecasterBehaviour, self.behaviour.current_behaviour
        )
        current_behaviour.setup()

        if ipfs_succeed:
            assert current_behaviour._forecasters is not None
        else:
            assert current_behaviour._forecasters is None

    def test_task_not_ready(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
    ) -> None:
        """Run test for behaviour when task result is not ready."""
        self._fast_forward(tmp_path)

        self.behaviour.context.task_manager.start()

        monkeypatch.setattr(AsyncResult, "ready", lambda *_: False)
        cast(
            TrainBehaviour, self.behaviour.current_behaviour
        ).params.sleep_time = SLEEP_TIME_TWEAK
        self.behaviour.act_wrapper()
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        assert (
            cast(
                APYEstimationBaseBehaviour, self.behaviour.current_behaviour
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )

    @pytest.mark.parametrize("ipfs_succeed", (True, False))
    def test_estimate_behaviour(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        ipfs_succeed: bool,
    ) -> None:
        """Run test for `EstimateBehaviour`."""
        self._fast_forward(tmp_path, ipfs_succeed)
        monkeypatch.setattr(TaskManager, "enqueue_task", lambda *_, **__: 0)
        monkeypatch.setattr(
            TaskManager, "get_task_result", lambda *_: DummyAsyncResult(pd.DataFrame())
        )

        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestCycleResetBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test CycleResetBehaviour."""

    behaviour_class = CycleResetBehaviour
    next_behaviour_class = FetchBatchBehaviour

    @pytest.mark.parametrize(
        "ipfs_succeed, log_level, log_message",
        (
            (True, logging.INFO, "Finalized estimates:"),
            (
                False,
                logging.ERROR,
                "There was an error while trying to fetch and load the estimations from IPFS!",
            ),
        ),
    )
    def test_reset_behaviour(
        self,
        monkeypatch: MonkeyPatch,
        tmp_path: PosixPath,
        caplog: LogCaptureFixture,
        no_action: Callable[[Any], None],
        ipfs_succeed: bool,
        log_level: int,
        log_message: str,
    ) -> None:
        """Test reset behaviour."""
        # Set data directory to a temporary path for tests.
        self.behaviour.context._agent_context._data_dir = tmp_path  # type: ignore

        # Send dummy forecasters to IPFS and get the hash.
        if ipfs_succeed:
            hash_ = cast(BaseBehaviour, self.behaviour.current_behaviour).send_to_ipfs(
                os.path.join(
                    tmp_path,
                    f"estimations_period_{self.synchronized_data.period_count}.csv",
                ),
                pd.DataFrame({"pool1": [1.435, 4.234], "pool2": [3.45, 23.64]}),
                filetype=SupportedFiletype.CSV,
            )
        else:
            hash_ = "non_existing"

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.behaviour_id,
            synchronized_data=SynchronizedData(
                AbciAppDB(setup_data=dict(most_voted_estimate=[hash_]))
            ),
        )
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.behaviour_class.behaviour_id

        monkeypatch.setattr(BenchmarkTool, "save", lambda _: no_action)
        monkeypatch.setattr(AbciApp, "last_timestamp", datetime.now())
        cast(
            CycleResetBehaviour, self.behaviour.current_behaviour
        ).params.observation_interval = SLEEP_TIME_TWEAK
        with caplog.at_level(
            log_level,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.act_wrapper()
        assert log_message in caplog.text
        time.sleep(SLEEP_TIME_TWEAK + 0.01)
        self.behaviour.act_wrapper()

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id

    def test_reset_behaviour_without_most_voted_estimate(
        self,
        monkeypatch: MonkeyPatch,
        no_action: Callable[[Any], None],
        caplog: LogCaptureFixture,
    ) -> None:
        """Test reset behaviour without most voted estimate."""
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.behaviour_id,
            synchronized_data=SynchronizedData(
                AbciAppDB(setup_data=dict(most_voted_estimate=[None]))
            ),
        )
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.behaviour_class.behaviour_id

        monkeypatch.setattr(BenchmarkTool, "save", lambda _: no_action)
        monkeypatch.setattr(AbciApp, "last_timestamp", datetime.now())

        self.behaviour.context.params.observation_interval = 0.1

        with caplog.at_level(
            logging.INFO,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.act_wrapper()
            cast(
                CycleResetBehaviour, self.behaviour.current_behaviour
            ).params.sleep_time = SLEEP_TIME_TWEAK
            time.sleep(SLEEP_TIME_TWEAK + 0.01)
            self.behaviour.act_wrapper()

        assert "[test_agent_name] Entered in the 'cycle_reset' behaviour" in caplog.text
        assert (
            "[test_agent_name] Finalized estimate not available. Resetting!"
            in caplog.text
        )

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestFreshModelResetBehaviour(APYEstimationFSMBehaviourBaseCase):
    """Test FreshModelResetBehaviour."""

    behaviour_class = FreshModelResetBehaviour
    next_behaviour_class = FetchBehaviour

    def test_fresh_model_reset_behaviour(self, caplog: LogCaptureFixture) -> None:
        """Run test for `ResetBehaviour`."""
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.behaviour_id,
            synchronized_data=SynchronizedData(AbciAppDB(setup_data={})),
        )
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.behaviour_class.behaviour_id

        with caplog.at_level(
            logging.INFO,
            logger="aea.test_agent_name.packages.valory.skills.apy_estimation_abci",
        ):
            self.behaviour.act_wrapper()

        assert (
            "[test_agent_name] Entered in the 'fresh_model_reset' behaviour"
            in caplog.text
        )
        assert (
            "[test_agent_name] Resetting to create a fresh forecasting model!"
            in caplog.text
        )

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint/types/params.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2

from packages.valory.connections.abci.gogoproto import (
    gogo_pb2 as gogoproto_dot_gogo__pb2,
)


DESCRIPTOR = _descriptor.FileDescriptor(
    name="tendermint/types/params.proto",
    package="tendermint.types",
    syntax="proto3",
    serialized_options=b"Z7github.com/tendermint/tendermint/proto/tendermint/types\250\342\036\001",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1dtendermint/types/params.proto\x12\x10tendermint.types\x1a\x14gogoproto/gogo.proto\x1a\x1egoogle/protobuf/duration.proto"\xf3\x01\n\x0f\x43onsensusParams\x12\x32\n\x05\x62lock\x18\x01 \x01(\x0b\x32\x1d.tendermint.types.BlockParamsB\x04\xc8\xde\x1f\x00\x12\x38\n\x08\x65vidence\x18\x02 \x01(\x0b\x32 .tendermint.types.EvidenceParamsB\x04\xc8\xde\x1f\x00\x12:\n\tvalidator\x18\x03 \x01(\x0b\x32!.tendermint.types.ValidatorParamsB\x04\xc8\xde\x1f\x00\x12\x36\n\x07version\x18\x04 \x01(\x0b\x32\x1f.tendermint.types.VersionParamsB\x04\xc8\xde\x1f\x00"G\n\x0b\x42lockParams\x12\x11\n\tmax_bytes\x18\x01 \x01(\x03\x12\x0f\n\x07max_gas\x18\x02 \x01(\x03\x12\x14\n\x0ctime_iota_ms\x18\x03 \x01(\x03"~\n\x0e\x45videnceParams\x12\x1a\n\x12max_age_num_blocks\x18\x01 \x01(\x03\x12=\n\x10max_age_duration\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationB\x08\xc8\xde\x1f\x00\x98\xdf\x1f\x01\x12\x11\n\tmax_bytes\x18\x03 \x01(\x03"2\n\x0fValidatorParams\x12\x15\n\rpub_key_types\x18\x01 \x03(\t:\x08\xb8\xa0\x1f\x01\xe8\xa0\x1f\x01".\n\rVersionParams\x12\x13\n\x0b\x61pp_version\x18\x01 \x01(\x04:\x08\xb8\xa0\x1f\x01\xe8\xa0\x1f\x01">\n\x0cHashedParams\x12\x17\n\x0f\x62lock_max_bytes\x18\x01 \x01(\x03\x12\x15\n\rblock_max_gas\x18\x02 \x01(\x03\x42=Z7github.com/tendermint/tendermint/proto/tendermint/types\xa8\xe2\x1e\x01\x62\x06proto3',
    dependencies=[
        gogoproto_dot_gogo__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_duration__pb2.DESCRIPTOR,
    ],
)


_CONSENSUSPARAMS = _descriptor.Descriptor(
    name="ConsensusParams",
    full_name="tendermint.types.ConsensusParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="block",
            full_name="tendermint.types.ConsensusParams.block",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=b"\310\336\037\000",
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="evidence",
            full_name="tendermint.types.ConsensusParams.evidence",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=b"\310\336\037\000",
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="validator",
            full_name="tendermint.types.ConsensusParams.validator",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=b"\310\336\037\000",
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="version",
            full_name="tendermint.types.ConsensusParams.version",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=b"\310\336\037\000",
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=106,
    serialized_end=349,
)


_BLOCKPARAMS = _descriptor.Descriptor(
    name="BlockParams",
    full_name="tendermint.types.BlockParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="max_bytes",
            full_name="tendermint.types.BlockParams.max_bytes",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="max_gas",
            full_name="tendermint.types.BlockParams.max_gas",
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="time_iota_ms",
            full_name="tendermint.types.BlockParams.time_iota_ms",
            index=2,
            number=3,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=351,
    serialized_end=422,
)


_EVIDENCEPARAMS = _descriptor.Descriptor(
    name="EvidenceParams",
    full_name="tendermint.types.EvidenceParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="max_age_num_blocks",
            full_name="tendermint.types.EvidenceParams.max_age_num_blocks",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="max_age_duration",
            full_name="tendermint.types.EvidenceParams.max_age_duration",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=b"\310\336\037\000\230\337\037\001",
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="max_bytes",
            full_name="tendermint.types.EvidenceParams.max_bytes",
            index=2,
            number=3,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=424,
    serialized_end=550,
)


_VALIDATORPARAMS = _descriptor.Descriptor(
    name="ValidatorParams",
    full_name="tendermint.types.ValidatorParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="pub_key_types",
            full_name="tendermint.types.ValidatorParams.pub_key_types",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"\270\240\037\001\350\240\037\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=552,
    serialized_end=602,
)


_VERSIONPARAMS = _descriptor.Descriptor(
    name="VersionParams",
    full_name="tendermint.types.VersionParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="app_version",
            full_name="tendermint.types.VersionParams.app_version",
            index=0,
            number=1,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"\270\240\037\001\350\240\037\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=604,
    serialized_end=650,
)


_HASHEDPARAMS = _descriptor.Descriptor(
    name="HashedParams",
    full_name="tendermint.types.HashedParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="block_max_bytes",
            full_name="tendermint.types.HashedParams.block_max_bytes",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="block_max_gas",
            full_name="tendermint.types.HashedParams.block_max_gas",
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=652,
    serialized_end=714,
)

_CONSENSUSPARAMS.fields_by_name["block"].message_type = _BLOCKPARAMS
_CONSENSUSPARAMS.fields_by_name["evidence"].message_type = _EVIDENCEPARAMS
_CONSENSUSPARAMS.fields_by_name["validator"].message_type = _VALIDATORPARAMS
_CONSENSUSPARAMS.fields_by_name["version"].message_type = _VERSIONPARAMS
_EVIDENCEPARAMS.fields_by_name[
    "max_age_duration"
].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
DESCRIPTOR.message_types_by_name["ConsensusParams"] = _CONSENSUSPARAMS
DESCRIPTOR.message_types_by_name["BlockParams"] = _BLOCKPARAMS
DESCRIPTOR.message_types_by_name["EvidenceParams"] = _EVIDENCEPARAMS
DESCRIPTOR.message_types_by_name["ValidatorParams"] = _VALIDATORPARAMS
DESCRIPTOR.message_types_by_name["VersionParams"] = _VERSIONPARAMS
DESCRIPTOR.message_types_by_name["HashedParams"] = _HASHEDPARAMS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ConsensusParams = _reflection.GeneratedProtocolMessageType(
    "ConsensusParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _CONSENSUSPARAMS,
        "__module__": "tendermint.types.params_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.types.ConsensusParams)
    },
)
_sym_db.RegisterMessage(ConsensusParams)

BlockParams = _reflection.GeneratedProtocolMessageType(
    "BlockParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _BLOCKPARAMS,
        "__module__": "tendermint.types.params_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.types.BlockParams)
    },
)
_sym_db.RegisterMessage(BlockParams)

EvidenceParams = _reflection.GeneratedProtocolMessageType(
    "EvidenceParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVIDENCEPARAMS,
        "__module__": "tendermint.types.params_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.types.EvidenceParams)
    },
)
_sym_db.RegisterMessage(EvidenceParams)

ValidatorParams = _reflection.GeneratedProtocolMessageType(
    "ValidatorParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _VALIDATORPARAMS,
        "__module__": "tendermint.types.params_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.types.ValidatorParams)
    },
)
_sym_db.RegisterMessage(ValidatorParams)

VersionParams = _reflection.GeneratedProtocolMessageType(
    "VersionParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _VERSIONPARAMS,
        "__module__": "tendermint.types.params_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.types.VersionParams)
    },
)
_sym_db.RegisterMessage(VersionParams)

HashedParams = _reflection.GeneratedProtocolMessageType(
    "HashedParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _HASHEDPARAMS,
        "__module__": "tendermint.types.params_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.types.HashedParams)
    },
)
_sym_db.RegisterMessage(HashedParams)


DESCRIPTOR._options = None
_CONSENSUSPARAMS.fields_by_name["block"]._options = None
_CONSENSUSPARAMS.fields_by_name["evidence"]._options = None
_CONSENSUSPARAMS.fields_by_name["validator"]._options = None
_CONSENSUSPARAMS.fields_by_name["version"]._options = None
_EVIDENCEPARAMS.fields_by_name["max_age_duration"]._options = None
_VALIDATORPARAMS._options = None
_VERSIONPARAMS._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the payloads.py module of the skill."""
import pytest

from packages.valory.skills.abstract_round_abci.base import Transaction
from packages.valory.skills.registration_abci.payloads import (
    RegistrationPayload,
    TransactionType,
)


def test_registration_abci_payload() -> None:
    """Test `RegistrationPayload`."""

    payload = RegistrationPayload(sender="sender")

    assert payload.initialisation is None
    assert payload.data == {}
    assert payload.transaction_type == TransactionType.REGISTRATION
    assert RegistrationPayload.from_json(payload.json) == payload


def test_registration_abci_payload_raises() -> None:
    """Test `RegistrationPayload`."""
    payload = RegistrationPayload(sender="sender", initialisation="0" * 10 ** 7)
    signature = "signature"
    tx = Transaction(payload, signature)
    with pytest.raises(ValueError):
        tx.encode()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the dialogues.py module of the skill."""
from enum import Enum
from typing import Type, cast
from unittest.mock import MagicMock

import pytest
from aea.protocols.dialogue.base import Dialogues
from aea.skills.base import Model

from packages.valory.skills.abstract_round_abci.dialogues import (
    AbciDialogue,
    AbciDialogues,
    ContractApiDialogue,
    ContractApiDialogues,
    HttpDialogue,
    HttpDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    SigningDialogue,
    SigningDialogues,
)


@pytest.mark.parametrize(
    "dialogues_cls,expected_role_from_first_message",
    [
        (AbciDialogues, AbciDialogue.Role.CLIENT),
        (HttpDialogues, HttpDialogue.Role.CLIENT),
        (SigningDialogues, SigningDialogue.Role.SKILL),
        (LedgerApiDialogues, LedgerApiDialogue.Role.AGENT),
        (ContractApiDialogues, ContractApiDialogue.Role.AGENT),
    ],
)
def test_dialogues_creation(
    dialogues_cls: Type[Model], expected_role_from_first_message: Enum
) -> None:
    """Test XDialogues creations."""
    dialogues = cast(Dialogues, dialogues_cls(name="", skill_context=MagicMock()))
    assert expected_role_from_first_message == dialogues._role_from_first_message(
        MagicMock(), MagicMock()
    )


def test_ledger_api_dialogue() -> None:
    """Test 'LedgerApiDialogue' creation."""
    dialogue = LedgerApiDialogue(MagicMock(), "", MagicMock())
    with pytest.raises(ValueError, match="Terms not set!"):
        dialogue.terms

    expected_terms = MagicMock()
    dialogue.terms = expected_terms
    assert expected_terms == dialogue.terms


def test_contract_api_dialogue() -> None:
    """Test 'ContractApiDialogue' creation."""
    dialogue = ContractApiDialogue(MagicMock(), "", MagicMock())
    with pytest.raises(ValueError, match="Terms not set!"):
        dialogue.terms

    expected_terms = MagicMock()
    dialogue.terms = expected_terms
    assert expected_terms == dialogue.terms



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Generates the specification for a given ABCI app in YAML/JSON/Mermaid format."""

import importlib
import json
import logging
import re
import sys
from itertools import product
from pathlib import Path
from typing import Any, Dict, List, OrderedDict, Set, TextIO, Tuple

import yaml


class DFASpecificationError(Exception):
    """Simple class to raise errors when parsing a DFA."""


class DFA:
    """Simple specification of a deterministic finite automaton (DFA)."""

    class OutputFormats:  # pylint: disable=too-few-public-methods
        """Output formats."""

        JSON = "json"
        YAML = "yaml"
        MERMAID = "mermaid"
        ALL = (JSON, YAML, MERMAID)

    def __init__(
        self,
        label: str,
        states: Set[str],
        default_start_state: str,
        start_states: Set[str],
        final_states: Set[str],
        alphabet_in: Set[str],
        transition_func: Dict[Tuple[str, str], str],
    ):  # pylint: disable=too-many-arguments
        """Initialize DFA object."""
        transition_func_in_states, transition_func_alphabet_in = map(
            set, zip(*transition_func.keys())
        )

        transition_func_states = transition_func_in_states.copy()
        transition_func_states.update(transition_func.values())  # type: ignore

        orphan_states = states - (start_states | set(transition_func.values()))

        error_strings = []
        if orphan_states:
            error_strings.append(
                f" - DFA spec. contains orphan states: {orphan_states}."
            )
        if not transition_func_states.issubset(states):
            error_strings.append(
                f" - Transition function contains unexpected states: {transition_func_states-states}."  # type: ignore
            )
        if not transition_func_alphabet_in.issubset(alphabet_in):
            error_strings.append(
                f" - Transition function contains unexpected input symbols: {transition_func_alphabet_in-alphabet_in}."  # type: ignore
            )
        if not alphabet_in.issubset(transition_func_alphabet_in):
            error_strings.append(
                f" - Unused input symbols: {alphabet_in-transition_func_alphabet_in}."  # type: ignore
            )
        if default_start_state not in start_states:
            error_strings.append(" - Default start state is not in start states set.")
        if not start_states.issubset(states):
            error_strings.append(
                f" - Start state set contains unexpected states: {start_states-states}."
            )
        if not final_states.issubset(states):
            error_strings.append(
                f" - Final state set contains unexpected states: {final_states-states}."
            )
        if start_states & final_states:
            error_strings.append(
                f" - Final state set contains start states: {start_states & final_states}."
            )
        if transition_func_in_states & final_states:
            error_strings.append(
                f" - Transitions out from final states: {transition_func_in_states & final_states}."
            )

        if len(error_strings) > 0:
            raise DFASpecificationError(
                "DFA spec. has the following issues:\n" + "\n".join(error_strings)
            )

        self.label = label
        self.states = states
        self.default_start_state = default_start_state
        self.start_states = start_states
        self.final_states = final_states
        self.alphabet_in = alphabet_in
        self.transition_func = transition_func

    def is_transition_func_total(self) -> bool:
        """
        Outputs True if the transition function of the DFA is total.

        A transition function is total when it explicitly defines all the transitions
        for all the possible pairs (state, input_symbol). By convention, when a transition
        (state, input_symbol) is not defined for a certain input_symbol, it will be
        automatically regarded as a self-transition to the same state.

        :return: None
        """
        return set(product(self.states, self.alphabet_in)) == set(
            self.transition_func.keys()
        )

    def get_transitions(self, input_sequence: List[str]) -> List[str]:
        """Runs the DFA given the input sequence of symbols, and outputs the list of state transitions."""
        state = self.default_start_state
        transitions = [state]
        for t in input_sequence:
            if t not in self.alphabet_in:
                logging.warning(
                    f"Input sequence contains a symbol {t} (ignored) not belonging to the DFA alphabet {self.alphabet_in}."
                )
            else:
                state = self.transition_func.get((state, t), state)
                transitions.append(state)
        return transitions

    def __eq__(self, other: object) -> bool:
        """Compares two DFAs"""
        if not isinstance(other, DFA):
            return NotImplemented  # Try reflected operation
        return self.__dict__ == other.__dict__

    def dump(self, file: Path, output_format: str = "yaml") -> None:
        """Dumps this DFA spec. to a file in YAML/JSON/Mermaid format."""

        dumper = getattr(self, f"dump_{output_format}")
        with open(file, "w+", encoding="utf-8") as fp:
            dumper(fp)

    def dump_json(self, fp: TextIO) -> None:
        """Dump to a json file."""
        json.dump(self.generate(), fp, indent=4)

    def dump_yaml(self, fp: TextIO) -> None:
        """Dump to a yaml file."""
        yaml.safe_dump(self.generate(), fp, indent=4)

    def dump_mermaid(self, fp: TextIO) -> None:
        """Dumps this DFA spec. to a file in Mermaid format."""
        print("stateDiagram-v2", file=fp)

        aux_map: Dict[Tuple[str, str], Set[str]] = {}
        for (s1, t), s2 in self.transition_func.items():
            aux_map.setdefault((s1, s2), set()).add(t)

        # A small optimization to make the output nicer:
        # (1) First, print the arrows that start from a start_state.
        for (s1, s2), t_set in aux_map.items():
            if s1 in self.start_states:
                print(
                    f"    {s1} --> {s2}: <center>{'<br />'.join(t_set)}</center>",
                    file=fp,
                )

        # (2) Then, print the rest of the arrows.
        for (s1, s2), t_set in aux_map.items():
            if s1 not in self.start_states:
                print(
                    f"    {s1} --> {s2}: <center>{'<br />'.join(t_set)}</center>",
                    file=fp,
                )

    def generate(self) -> Dict[str, Any]:
        """Retrieves an exportable respresentation for YAML/JSON dump of this DFA."""
        dfa_export: Dict[str, Any] = {}
        for k, v in self.__dict__.items():
            if isinstance(v, Set):
                dfa_export[k] = sorted(v)
            elif isinstance(v, Dict):
                dfa_export[k] = dict(
                    OrderedDict(
                        [(str(k2).replace("'", ""), v2) for k2, v2 in v.items()]
                    )
                )
            else:
                dfa_export[k] = v
        return dfa_export

    @classmethod
    def _norep_list_to_set(cls, lst: List[str]) -> Set[str]:
        """Converts a list to a set and ensures that it does not contain repetitions."""
        if not isinstance(lst, List):
            raise DFASpecificationError(f"DFA spec. object {lst} is not of type List.")
        if len(lst) != len(set(lst)):
            raise DFASpecificationError(
                f"DFA spec. List {lst} contains repeated values."
            )
        return set(lst)

    @classmethod
    def _str_to_tuple(cls, k: str) -> Tuple[str, str]:
        """Converts a string in format "(a, b)" to a tuple ("a", "b")."""
        match = re.search(r"\((\w*),\s(\w*)\)", k, re.DOTALL)

        if match is None:
            raise DFASpecificationError(
                f"DFA spec. JSON file contains an invalid transition function key: {k}."
            )

        return (match.group(1), match.group(2))

    @classmethod
    def load(cls, fp: TextIO, input_format: str = "yaml") -> "DFA":
        """Loads a DFA JSON specification from file."""

        if input_format == "json":
            dfa_simple = json.load(fp)
        elif input_format == "yaml":
            dfa_simple = yaml.safe_load(fp)
        else:
            raise ValueError(f"Unrecognized input format {input_format}.")

        try:
            label = dfa_simple.pop("label")
            states = DFA._norep_list_to_set(dfa_simple.pop("states"))
            default_start_state = dfa_simple.pop("default_start_state")
            start_states = DFA._norep_list_to_set(dfa_simple.pop("start_states"))
            final_states = DFA._norep_list_to_set(dfa_simple.pop("final_states"))
            alphabet_in = DFA._norep_list_to_set(dfa_simple.pop("alphabet_in"))
            transition_func = {
                DFA._str_to_tuple(k): v
                for k, v in dfa_simple.pop("transition_func").items()
            }
        except KeyError as ke:
            raise DFASpecificationError("DFA spec. JSON file missing key.") from ke

        if len(dfa_simple) != 0:
            raise DFASpecificationError(
                f"DFA spec. JSON file contains unexpected objects: {dfa_simple.keys()}."
            )

        return cls(
            label,
            states,
            default_start_state,
            start_states,
            final_states,
            alphabet_in,
            transition_func,
        )

    @classmethod
    def abci_to_dfa(cls, abci_app_cls: Any, label: str = "") -> "DFA":
        """Translates an AbciApp class into a DFA."""

        trf = abci_app_cls.transition_function

        label = (
            label
            if label != ""
            else abci_app_cls.__module__ + "." + abci_app_cls.__name__
        )
        default_start_state = abci_app_cls.initial_round_cls.__name__
        start_states = DFA._norep_list_to_set(
            [s.__name__ for s in abci_app_cls.initial_states]
        )
        start_states = start_states if start_states else {default_start_state}
        final_states = DFA._norep_list_to_set(
            [s.__name__ for s in abci_app_cls.final_states]
        )
        states = (
            DFA._norep_list_to_set([k.__name__ for k in trf])
            | {s.__name__ for k in trf for s in trf[k].values()}
            | start_states
            | final_states
        )
        alphabet_in = {str(s).rsplit(".", 1)[1] for k in trf for s in trf[k].keys()}
        transition_func = {
            (k.__name__, str(s).rsplit(".", 1)[1]): trf[k][s].__name__
            for k in trf
            for s in trf[k]
        }
        transition_func = OrderedDict(sorted(transition_func.items()))
        return cls(
            label,
            states,
            default_start_state,
            start_states,
            final_states,
            alphabet_in,
            transition_func,
        )


class SpecCheck:
    """Class to represent abci spec checks."""

    @staticmethod
    def check_one(informat: str, infile: str, classfqn: str) -> bool:
        """Check for one."""

        print(f"Checking : {classfqn}")
        module_name, class_name = classfqn.rsplit(".", 1)
        module = importlib.import_module(module_name)
        if not hasattr(module, class_name):
            raise Exception(f'Class "{class_name}" is not in "{module_name}".')

        abci_app_cls = getattr(module, class_name)
        dfa1 = DFA.abci_to_dfa(abci_app_cls, classfqn)

        with open(infile, "r", encoding="utf-8") as fp:
            dfa2 = DFA.load(fp, informat)

        return dfa1 == dfa2

    @classmethod
    def check_all(cls, packages_dir: Path) -> None:
        """Check all the available definitions."""

        did_not_match = []
        fsm_specifications = sorted(
            [
                *packages_dir.glob("**/fsm_specification.yaml"),
                *packages_dir.glob("**/fsm_specification_composition.yaml"),
            ]
        )
        for spec_file in fsm_specifications:
            with open(str(spec_file), mode="r", encoding="utf-8") as fp:
                specs = yaml.safe_load(fp)
                if not cls.check_one(
                    informat="yaml", infile=str(spec_file), classfqn=specs.get("label")
                ):
                    did_not_match.append(f"{specs.get('label')} - {str(spec_file)}")

        if len(did_not_match) > 0:
            print("\nSpecifications did not match for following definitions.\n")
            print("\n".join(did_not_match))
            sys.exit(1)

        print("Check successful.")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the payloads of the skill."""
from typing import Optional

import pytest

from packages.valory.skills.transaction_settlement_abci.payloads import (
    CheckTransactionHistoryPayload,
    FinalizationTxPayload,
    RandomnessPayload,
    ResetPayload,
    SelectKeeperPayload,
    SignaturePayload,
    SynchronizeLateMessagesPayload,
    TransactionType,
    ValidatePayload,
)


def test_randomness_payload() -> None:
    """Test `RandomnessPayload`."""

    payload = RandomnessPayload(sender="sender", round_id=1, randomness="test")

    assert payload.round_id == 1
    assert payload.randomness == "test"
    assert payload.data == {"round_id": 1, "randomness": "test"}
    assert payload.transaction_type == TransactionType.RANDOMNESS


def test_select_keeper_payload() -> None:
    """Test `SelectKeeperPayload`."""

    payload = SelectKeeperPayload(sender="sender", keepers="test")

    assert payload.keepers == "test"
    assert payload.data == {"keepers": "test"}
    assert payload.transaction_type == TransactionType.SELECT_KEEPER


@pytest.mark.parametrize("vote", (None, True, False))
def test_validate_payload(vote: Optional[bool]) -> None:
    """Test `ValidatePayload`."""

    payload = ValidatePayload(sender="sender", vote=vote)

    assert payload.vote is vote
    assert payload.data == {} if vote is None else {"vote": vote}
    assert payload.transaction_type == TransactionType.VALIDATE


def test_tx_history_payload() -> None:
    """Test `CheckTransactionHistoryPayload`."""

    payload = CheckTransactionHistoryPayload(sender="sender", verified_res="test")

    assert payload.verified_res == "test"
    assert payload.data == {"verified_res": "test"}
    assert payload.transaction_type == TransactionType.CHECK


def test_synchronize_payload() -> None:
    """Test `SynchronizeLateMessagesPayload`."""

    payload = SynchronizeLateMessagesPayload(sender="sender", tx_hashes="test")

    assert payload.tx_hashes == "test"
    assert payload.data == {"tx_hashes": "test"}
    assert payload.transaction_type == TransactionType.SYNCHRONIZE


def test_signature_payload() -> None:
    """Test `SignaturePayload`."""

    payload = SignaturePayload(sender="sender", signature="sign")

    assert payload.signature == "sign"
    assert payload.data == {"signature": "sign"}
    assert payload.transaction_type == TransactionType.SIGNATURE


def test_finalization_tx_payload() -> None:
    """Test `FinalizationTxPayload`."""

    payload = FinalizationTxPayload(
        sender="sender",
        tx_data={
            "tx_digest": "hash",
            "nonce": 0,
            "max_fee_per_gas": 0,
            "max_priority_fee_per_gas": 0,
        },
    )

    assert payload.data == {
        "tx_data": {
            "tx_digest": "hash",
            "nonce": 0,
            "max_fee_per_gas": 0,
            "max_priority_fee_per_gas": 0,
        }
    }
    assert payload.transaction_type == TransactionType.FINALIZATION


def test_reset_payload() -> None:
    """Test `ResetPayload`."""

    payload = ResetPayload(sender="sender", period_count=1)

    assert payload.period_count == 1
    assert payload.data == {"period_count": 1}
    assert payload.transaction_type == TransactionType.RESET
    assert ResetPayload.from_json(payload.json) == payload



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the behaviours for the 'counter_client' skill."""
# isort:skip_file  # noqa
import json
from abc import ABC
from typing import Dict, Optional, cast

from aea.skills.base import Behaviour
from aea.skills.behaviours import TickerBehaviour

from packages.valory.connections.http_client.connection import (  # type: ignore # pylint: disable=no-name-in-module,import-error
    PUBLIC_ID as HTTP_CLIENT_PUBLIC_ID,  # type: ignore
)
from packages.valory.protocols.http import HttpMessage  # type: ignore # pylint: disable=no-name-in-module,import-error

# type: ignore # pylint: disable=no-name-in-module,import-error
from packages.valory.skills.counter_client.dialogues import HttpDialogues
from packages.valory.skills.counter_client.handlers import curdatetime


class BaseBehaviour(Behaviour, ABC):
    """Abstract base behaviour for this skill."""

    is_programmatically_defined = True

    @property
    def tendermint_url(self) -> str:
        """Get the Tendermint url."""
        return cast(str, "http://" + self.context.params.tendermint_url)

    def send_http_request_message(
        self,
        method: str,
        url: str,
        content: Dict = None,
        handler_callback: Optional[str] = None,
    ) -> None:
        """
        Send an http request message.

        :param method: the http request method (i.e. 'GET' or 'POST').
        :param url: the url to send the message to.
        :param content: the payload.
        :param handler_callback: the handler callback. It must match
            the name of the method to be called in the handler.
        """
        # context
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)

        # http request message
        request_http_message, http_dialogue = http_dialogues.create(
            counterparty=str(HTTP_CLIENT_PUBLIC_ID),
            performative=HttpMessage.Performative.REQUEST,
            method=method,
            url=url,
            headers="",
            version="",
            body=b""
            if content is None
            else json.dumps(content, sort_keys=True).encode("utf-8"),
        )
        # send
        self.context.outbox.put_message(message=request_http_message)
        request_nonce = http_dialogue.dialogue_label.dialogue_reference[0]
        self.context.state.request_to_handler[request_nonce] = handler_callback


class MonitorBehaviour(TickerBehaviour, BaseBehaviour):
    """Send an ABCI query periodically."""

    def query(self) -> None:
        """Send a query request."""
        self.send_http_request_message(
            "GET", self.tendermint_url + "/abci_query", handler_callback="abci_query"
        )

    def act(self) -> None:
        """Do the action."""
        self.query()

    def setup(self) -> None:
        """Set up the behaviour."""

    def teardown(self) -> None:
        """Tear down the behaviour."""


class IncrementerBehaviour(TickerBehaviour, BaseBehaviour):
    """Send a transaction"""

    def send_transaction(self) -> None:
        """Send a query request."""
        next_count = self.context.state.count + 1
        self.context.logger.info(
            f"[{curdatetime()}] Sending transaction with new count: {next_count}"
        )
        self.send_http_request_message(
            "GET",
            self.tendermint_url + "/broadcast_tx_commit?tx=0x{:02x}".format(next_count),
            handler_callback="broadcast_tx_commit",
        )

    def act(self) -> None:
        """Do the action."""
        self.send_transaction()

    def setup(self) -> None:
        """Set up the behaviour."""

    def teardown(self) -> None:
        """Tear down the behaviour."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for the Tendermint com server."""

import logging
import os
import shutil
import socket
import stat
import subprocess  # nosec
import tempfile
import time
import urllib
from pathlib import Path
from typing import Any, Callable, List

import flask
import requests

from deployments.Dockerfiles.localnode.app import (  # type: ignore
    CONFIG_OVERRIDE,
    create_app,
    get_defaults,
    load_genesis,
    override_config_toml,
)
from deployments.Dockerfiles.localnode.tendermint import (  # type: ignore
    DEFAULT_RPC_LISTEN_ADDRESS,
    TendermintNode,
)


ENCODING = "utf-8"
VERSION = "0.34.11"
HTTP = "http://"

parse_result = urllib.parse.urlparse(DEFAULT_RPC_LISTEN_ADDRESS)  # type: ignore
IP, PORT = parse_result.hostname, parse_result.port


# utility functions
def readonly_handler(func: Callable, path: str, execinfo) -> None:  # type: ignore
    """If permission is readonly, we change and retry."""
    os.chmod(path, stat.S_IWRITE)
    func(path)


def port_is_open(ip: str, port: int) -> bool:
    """Assess whether a port is open"""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    is_open = sock.connect_ex((ip, port)) == 0
    sock.close()
    return is_open


def wait_for_node_to_run(func: Callable) -> Callable:
    """Wait for Tendermint node to run"""

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        i, max_retries = 0, 5
        while not port_is_open(IP, PORT) and i < max_retries:
            logging.debug(f"waiting for node... t={i}")
            i += 1
            time.sleep(1)
        response = requests.get(f"{HTTP}{IP}:{PORT}/status")
        success = response.status_code == 200
        assert success, "Tendermint node not running"
        func(*args, **kwargs)

    return wrapper


# base classes
class BaseTendermintTest:
    """BaseTendermintTest"""

    tendermint: str
    tm_home: str
    path: Path

    @classmethod
    def setup_class(cls) -> None:
        """Setup the test."""
        cls.tendermint = shutil.which("tendermint")  # type: ignore
        cls.tm_home = os.environ["TMHOME"] = tempfile.mkdtemp()
        cls.path = Path(cls.tm_home)
        assert not os.listdir(cls.tm_home)

        command = [cls.tendermint, "init", "validator", "--home", cls.tm_home]
        process = subprocess.Popen(command, stderr=subprocess.PIPE)  # nosec
        _, stderr = process.communicate()
        assert not stderr, stderr

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown the test."""
        shutil.rmtree(cls.tm_home, ignore_errors=True, onerror=readonly_handler)


class BaseTendermintServerTest(BaseTendermintTest):
    """Test Tendermint server app"""

    app: flask.app.Flask
    app_context: flask.ctx.AppContext
    tendermint_node: TendermintNode

    @classmethod
    def setup_class(cls) -> None:
        """Setup the test."""
        super().setup_class()
        os.environ["PROXY_APP"] = "kvstore"
        os.environ["CREATE_EMPTY_BLOCKS"] = "true"
        os.environ["LOG_FILE"] = str(cls.path / "tendermint.log")
        cls.app, cls.tendermint_node = create_app(cls.path / "tm_state")
        cls.app.config["TESTING"] = True
        cls.app_context = cls.app.app_context()
        cls.app_context.push()

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown the test."""
        cls.app_context.pop()
        cls.tendermint_node.stop()
        super().teardown_class()


# unit tests
def test_tendermint_executable_found() -> None:
    """Test Tendermint executable found"""
    assert shutil.which("tendermint"), "No `tendermint` executable found"
    output = subprocess.check_output(["tendermint", "version"])  # nosec
    assert output.decode(ENCODING).strip() == VERSION


class TestTendermintServerUtilityFunctions(BaseTendermintTest):
    """Test Tendermint server utility functions"""

    def test_load_genesis(self) -> None:
        """Load the genesis.json file"""
        expected_keys = {
            "genesis_time",
            "chain_id",
            "initial_height",
            "consensus_params",
            "validators",
            "app_hash",
        }
        json = load_genesis()
        assert set(json.keys()) == expected_keys

    def test_get_defaults(self) -> None:
        """Test get_defaults function"""
        expected_keys = {"genesis_time"}
        defaults = get_defaults()
        assert set(defaults.keys()) == expected_keys

    def test_override_config_toml(self) -> None:
        """Test override_config_toml function"""
        path = self.path / "config" / "config.toml"
        assert path.is_file(), f"not a file: {path}"
        txt = path.read_text("utf-8")
        assert all(old in txt and new not in txt for old, new in CONFIG_OVERRIDE)
        override_config_toml()
        txt = path.read_text("utf-8")
        assert not any(old in txt or new not in txt for old, new in CONFIG_OVERRIDE)


# integration tests
class TestTendermintServerApp(BaseTendermintServerTest):
    """Test Tendermint server app"""

    @wait_for_node_to_run
    def test_files_exist(self) -> None:
        """Test that the necessary files are present"""

        def remove_prefix(text: str, prefix: str) -> str:
            """str.removeprefix only from python3.9 onward"""
            return text[text.startswith(prefix) and len(prefix) :]

        expected_file_names = [
            "/config/config.toml",
            "/config/priv_validator_key.json",
            "/config/genesis.json",
            "/config/node_key.json",
            "/data/priv_validator_state.json",
        ]

        file_names = [remove_prefix(str(p), self.tm_home) for p in self.path.rglob("*")]
        # we may create extra files, just want to check these exist
        missing_files = set(expected_file_names).difference(file_names)
        assert not missing_files

    @wait_for_node_to_run
    def test_get_request_status(self) -> None:
        """Check local node is running"""
        response = requests.get(f"{HTTP}{IP}:{PORT}/status")
        data = response.json()
        assert data["result"]["node_info"]["version"] == VERSION

    @wait_for_node_to_run
    def test_handle_notfound(self) -> None:
        """Test handle not found"""
        with self.app.test_client() as client:
            response = client.get("/non_existing_endpoint")
            assert response.status_code == 404

    @wait_for_node_to_run
    def test_get_app_hash(self) -> None:
        """Test get app hash"""
        time.sleep(3)  # requires some extra time!
        with self.app.test_client() as client:
            response = client.get("/app_hash")
            data = response.get_json()
            assert response.status_code == 200
            assert "error" not in data
            assert "app_hash" in data


class TestTendermintGentleResetServer(BaseTendermintServerTest):
    """Test Tendermint gentle reset"""

    @wait_for_node_to_run
    def test_gentle_reset(self) -> None:
        """Test gentle reset"""
        with self.app.test_client() as client:
            response = client.get("/gentle_reset")
            data = response.get_json()
            assert response.status_code == 200
            assert data["status"] is True


class TestTendermintHardResetServer(BaseTendermintServerTest):
    """Test Tendermint hard reset"""

    @wait_for_node_to_run
    def test_hard_reset(self) -> None:
        """Test hard reset"""
        with self.app.test_client() as client:
            response = client.get("/hard_reset")
            data = response.get_json()
            assert response.status_code == 200
            assert data["status"] is True


class TestTendermintLogMessages(BaseTendermintServerTest):
    """Test Tendermint message logging"""

    @wait_for_node_to_run
    def test_tendermint_logs(self) -> None:
        """Test Tendermint logs"""

        def get_logs() -> str:
            with open(os.environ["LOG_FILE"], "r") as f:
                lines = "".join(f.readlines())
            return lines

        def get_missing(messages: List[str]) -> List[str]:
            i, max_retries = 0, 5
            while messages and i < max_retries:
                i += 1
                time.sleep(1)
                messages = [line for line in messages if line not in get_logs()]
            return messages

        before_stopping = [
            "Tendermint process started",
            "Monitoring thread started",
            "Starting multiAppConn service",
            "Starting localClient service",
            "This node is a validator",
        ]

        after_stopping = [
            "Monitoring thread terminated",
            "Tendermint process stopped",
        ]

        assert not get_missing(before_stopping)
        self.tendermint_node.stop()
        assert not get_missing(after_stopping)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""BaseChannel for MockNode"""
from typing import Dict

import packages.valory.connections.abci.tendermint.abci.types_pb2 as abci_types  # type: ignore


class BaseChannel:
    """Defines the interface for other channels to use"""

    def __init__(self, **kwargs: Dict) -> None:
        """Initializes a channel"""
        pass

    def connect(self) -> None:
        """
        Set up the channel.

        By default, it is a no-op.
        """

    def disconnect(self) -> None:
        """
        Tear down the channel.

        By default, it is a no-op.
        """

    def send_info(self, request: abci_types.RequestInfo) -> abci_types.ResponseInfo:
        """
        Sends an info request.

        :param: request: RequestInfo pb object
        """
        raise NotImplementedError

    def send_echo(self, request: abci_types.RequestEcho) -> abci_types.ResponseEcho:
        """
        Sends an echo request.

        :param: request: RequestEcho pb object
        """
        raise NotImplementedError

    def send_flush(self, request: abci_types.RequestFlush) -> abci_types.ResponseFlush:
        """
        Sends an flush request.

        :param: request: RequestFlush pb object
        """
        raise NotImplementedError

    def send_set_option(
        self, request: abci_types.RequestSetOption
    ) -> abci_types.ResponseSetOption:
        """
        Sends an setOption request.

        :param: request: RequestSetOption pb object
        """
        raise NotImplementedError

    def send_deliver_tx(
        self, request: abci_types.RequestDeliverTx
    ) -> abci_types.ResponseDeliverTx:
        """
        Sends an deliverTx request.

        :param: request: RequestDeliverTx pb object
        """
        raise NotImplementedError

    def send_check_tx(
        self, request: abci_types.RequestCheckTx
    ) -> abci_types.ResponseCheckTx:
        """
        Sends an checkTx request.

        :param: request: RequestCheckTx pb object
        """
        raise NotImplementedError

    def send_query(self, request: abci_types.RequestQuery) -> abci_types.ResponseQuery:
        """
        Sends an query request.

        :param: request: RequestQuery pb object
        """
        raise NotImplementedError

    def send_commit(
        self, request: abci_types.RequestCommit
    ) -> abci_types.ResponseCommit:
        """
        Sends an commit request.

        :param: request: RequestCommit pb object
        """
        raise NotImplementedError

    def send_init_chain(
        self, request: abci_types.RequestInitChain
    ) -> abci_types.ResponseInitChain:
        """
        Sends an initChain request.

        :param: request: RequestInitChain pb object
        """
        raise NotImplementedError

    def send_begin_block(
        self, request: abci_types.RequestBeginBlock
    ) -> abci_types.ResponseBeginBlock:
        """
        Sends an beginBlock request.

        :param: request: RequestBeginBlock pb object
        """
        raise NotImplementedError

    def send_end_block(
        self, request: abci_types.RequestEndBlock
    ) -> abci_types.ResponseEndBlock:
        """
        Sends an endBlock request.

        :param: request: RequestEndBlock pb object
        """
        raise NotImplementedError

    def send_list_snapshots(
        self, request: abci_types.RequestListSnapshots
    ) -> abci_types.ResponseListSnapshots:
        """
        Sends an listSnapshots request.

        :param: request: RequestListSnapshots pb object
        """
        raise NotImplementedError

    def send_offer_snapshot(
        self, request: abci_types.RequestOfferSnapshot
    ) -> abci_types.ResponseOfferSnapshot:
        """
        Sends an offerSnapshot request.

        :param: request: RequestOfferSnapshot pb object
        """
        raise NotImplementedError

    def send_load_snapshot_chunk(
        self, request: abci_types.RequestLoadSnapshotChunk
    ) -> abci_types.ResponseLoadSnapshotChunk:
        """
        Sends an loadSnapshotChunk request.

        :param: request: RequestLoadSnapshotChunk pb object
        """
        raise NotImplementedError

    def send_apply_snapshot_chunk(
        self, request: abci_types.RequestApplySnapshotChunk
    ) -> abci_types.ResponseApplySnapshotChunk:
        """
        Sends an applySnapshotChunk request.

        :param: request: RequestApplySnapshotChunk pb object
        """
        raise NotImplementedError
